var g_version = "v0.4.4";

var pieceLetters = {
	englishN: ["P", "N", "B", "R", "Q", "K", "H", "U",
	           "F",  "W",  "M",  "A",  "C",  "V",  "D", "X"],

	englishS: ["P", "S", "B", "R", "Q", "K", "N", "BP",
	           "FE", "WE", "EK", "PR", "EM", "AM", "?", "X"],

	french:   ["P", "C", "F", "T", "D", "R", "N", "PB",
	           "FE", "VI", "RE", "PR", "IM", "AM", "?", "X"],

	german:   ["B", "S", "L", "T", "D", "K", "N", "BB",
	           "FE", "WE", "EK", "PR", "KN", "AM", "?", "X"]

//	germanPDB: ["B", "S", "L", "T", "D", "K", "N", "BE",
//	            "FE", "WE", "EK", "PR", "KA", "AM", "?", "X"]
};

var pieceFigurines = {
	w: ["&#9817;", "&#9816;", "&#9815;", "&#9814;", "&#9813;", "&#9812;",
	    "wN", "wBP",
	    "wFE", "wWE", "wEK", "wPR", "wEM", "wAM", "w?", "&#9675;"],
	b: ["&#9823;", "&#9822;", "&#9821;", "&#9820;", "&#9819;", "&#9818;",
	    "bN", "bBP",
	    "bFE", "bWE", "bEK", "bPR", "bEM", "bAM", "b?", "&#9679;"]
};

var commandNames = [
	// Language commands.
	"BeginProblem: englishS",
	"DebutProbleme: french", "D\u00E9butProbl\u00E8me: french",
	"AnfangProblem: german",
	"EnglishN: englishN",

	// Commands to ignore.
	"EndProblem: ignore", "FinProbleme: ignore", "EndeProblem: ignore",

	// Commands that are comments.
	"remark: comment", "remarque: comment", "Bemerkung: comment",
	"author: comment", "auteur: comment", "Autor: comment",
	"origin: comment", "source: comment", "Quelle: comment",
	"title: comment", "titre: comment", "Titel: comment",

	// Commands that do something.
	"forsyth",
	"pieces", "pi\u00E8ces: pieces", "Steine: pieces",
	"option",
	"play",
	"test",
	"set",
	"stipulation", "enonce: stipulation", "\u00E9nonc\u00E9: stipulation",
	"Forderung: stipulation",

	"condition", "Bedingung: condition"
];

// Make these language-dependent because "black" and "blanc" share a long
// prefix. [0] and [1] are the preferred white and black names. Others are
// alternatives.
var pieceColorNames = {
	englishN: ["white", "black"],
	englishS: ["white", "black"],
	french: ["blanc: white", "noir: black", "blancs: white", "noirs: black"],
	german: ["weiss: white", "schwarz: black", "wei\u00DF: white"]
};

/* TODO
var pieceAttributeNames = {
	englishN: ["royal"],
	englishS: ["royal"],
	french: ["Royales: royal"],
	german: ["Koeniglich: royal"]
};
*/

var optionNames = [
	"FirstToMove",
	"InitialMoveNumber"
];

// The numbering needs to be kept in sync with the C engine.
var playNames = [
	"help: 0",
	"defense: 1", "defence: 1", "defensive: 1",
	"series: 2",
	"pass: 3"
];

// The numbering needs to be kept in sync with the C engine.
// We can't have anything which starts with h, *, s or r.
var testNames = [
	"not",	// negates the test
	"dia: 0", "a=>b: 0", "pg: 0", "pj: 0", "bp: 0",	// diagram
	"+: 1",	// check
	"#: 2",	// mate
	"=: 3",	// stalemate
	"!=: 4", // auto-stalemate
	"==: 5", 	// double-stalemate
	"!+: 6", // auto-check
	"!#: 7", 	// auto-mate
	"##: 8", 	// double-mate
	"x: 9",	// capture
	"%: 10",	// win a piece
	"00: 11", "OO: 11", "-00: 11", "-OO: 11",	// castling
	"ep: 12", "-ep: 12",	// en passant
	"2+: 13", "++: 13",	// double check
	"3+: 14", "+++: 14",	// triple check
	"4+: 15", "++++: 15",	// quadruple check
	"5+: 16", "+++++: 16",	// fivefold check
	"6+: 17", "++++++: 17",	// sixfold check
	"7+: 18", "+++++++: 18",	// sevenfold check
	"8+: 19", "++++++++: 19",	// eightfold check
	"~: 20"	// any move
//	"%",	// win-a-piece
//	"#=",	// mate or stalemate
//	http://bulletin.chessproblems.ca/pdf/cpb-02.pdf
//	"!=" // auto-stalemate (become stalemated)
//  "Z" // target-square
//  "ct" // "RK" circuit (a piece moves and come back to the square it occupied in the diagram position.)
//  "<>" // "PW" place-exchange (two pieces switch places)
//  "F" // pin (some non-king move is illegal because of self-check?)
//  "!F" // auto-pin (pins own unit)
//	http://www.jurajlorinc.com/chess/explanat.htm
//	"##"	// beidmatt: it is the position in which both kings are mated. The last move reaching this position may be played even if it leaves own king under check or it is selfcheck, but the king of the side playing the move reaching beidmatt mustn't be mated yet before this move is played
//	"##!"	// gegenmatt: beidmatt except that the king of the side playing the move reaching gegenmatt must be mated before that move. Gegenmatt is usually called doublemate in English, beidmatt having no counterpart. 

/*

begin
forsyth 8/8/R1K1k2r/8/8/1rP5/2r5/5R2
stip h##!1.0
end

  1.Rc2*c3 + Kc6-d6 ##!

  1.Rb3*c3 + Kc6-d6 ##

*/

];

// The numbering needs to be kept in sync with the C engine.
// TODO: use translated color names.

var setNames = [
	"test: 1",
	"pass: 2"
];

var setOptions = [
	// Options for "set test"
	"first: 1",
	"second: 2",
	"white: 3",
	"black: 4",

	"accept: 101",
	"reject: 102",
	"prefer: 103",
	"avoid: 104",
	"only: 105",
	"none: 106",
	"one: 107",	// Accept move only if it is the only one working.

	"wake",		// Wake idle side if special moves exist.
	"end",

	// Convenience options.
	"reflex",	// Shorthand for "prefer wake end".
	"living",	// Cancels the "end" option.

	// Option for "set pass".
	"unique",	// Accept parry move only if it is unique, similar
	         	// to "defense" processing.

	// Applies to both.
	"pure"		// Applies only to play type "series".
];

var prefixNames = [
	// fixed length with no closure
	"",		// for non-dia, must have length 0.5, equivalent to h0.5
	   		// for dia, equivalent to h.
	"s",	// must have length 1.0, equivalent to hs1.0
	"r",	// must have length 1.0, equivalent to hr1.0

	// help
	"h",
	"ph",
	"ah",
	"pah",

	// series
	"ser-",
	"phser-",
	"ahser-",
	"pahser-"
];

// Closure consists of any sequence of h and * and optionally end with s or r.

var diagramNames = [
	"forsyth",
	"pieces", "Steine: pieces"
];

var diagramOptions = [
	"AddPieces: AddWhitePieces AddBlackPieces",
		"AjouterDesPieces: AddWhitePieces AddBlackPieces",
		"ErganzeSteine: AddWhitePieces AddBlackPieces",
	"AddWhitePieces",
		"AjouterDesPiecesBlanches: AddWhitePieces",
		"ErganzeWeisseSteine: AddWhitePieces",
	"AddBlackPieces",
		"AjouterDesPiecesNoires: AddBlackPieces",
		"ErganzeSchwarzeSteine: AddBlackPieces",
	"ColorThePieces",
		"ColorierLesPieces: ColorThePieces",
		"FarbeDieSteine: ColorThePieces"
];

var conditionNames = [
	"Chess960",
	"Chigorin",
	"Almost",
	"Knightmate",
	"CavalierMajeur",
	"Berolina",
	"NonRoyalKings",
	"Extinction",
	"RoyalDynasty",

	"Annan",
	"BackToBack",
	"FaceToFace",
	"VerticalCylinder",
	"VerticalCylinder_NoExtendedCastling",
	"Monochromatic",
	"Monochromatic_Anti",
	"Grid",
	"WhiteEdgemover",
	"BlackEdgemover",

	"Madrasi",
	"Madrasi_RexInclusive",
	"Eiffel",

	"Woozles",
	"Woozles_RexExclusive",
	"BiWoozles",
	"BiWoozles_RexExclusive",

	"Patrol",
	"Patrol_Ultra",
	"Patrol_Anti",
	"Provocation",
	"Provocation_Ultra",
	"Provocation_Anti",
	"Geneva",
	"Geneva_RexInclusive",
	"SuperGuards",
	"Immun",
	"Immun_RexInclusive",
	"Disparate",

	"FreeForAll",
	"Bicapture",
	"Bicapture_RexExclusive",
	"Cannibal",
	"NormalPawn",

	"Glasgow",
	"PromOnly",
	"PromOnly_P",
	"PromOnly_S",
	"PromOnly_B",
	"PromOnly_R",
	"PromOnly_Q",
	"PromOnly_K",
	"AdversePromotions",
	"Multicapture",
	"Multicapture_Anti",

	"Chameleon",
	"Chameleon_Reversal",
	"Relegation",
	"Einstein",
	"Einstein_Reversal",
	"Einstein_Anti",

	"Andernach",
	"AntiAndernach",
	"Tibet",
	"Tibet_Double",
	"Volage",
	"Volage_Hyper",

	"KobulKings",

	"Circe",
	"Circe_RexInclusive",
	"Circe_Chameleon",
	"Circe_Clone",
	"Circe_Equipollents",
	"Circe_Parrain",
	"Circe_File",
	"Circe_Symmetry",
	"Circe_Diagram",
	"Circe_Platzwechsel",
	"Circe_Antipodes",
	"Circe_Mirror",
	"Circe_Couscous",
	"Circe_Diametral",
	"Circe_VerticalMirror",
	"Circe_Strict",
	"Circe_Assassin",
	"Circe_Turncoats",

	"Sentinels",
	"Sentinels_PionAdverse",
	"Sentinels_Berolina",
	"Sentinels_Para",
	"Sentinels_Totalmax16",
	"Masand",

	"NoImmediateRepetition",
	"Koeko",
	"Koeko_Anti",
	"Koeko_New",
	"Koeko_AffectedByGrid",
	"Koeko_AppliesToRebirth",
	"Isardam",
	"Isardam_TypeB",
	"Isardam_RexInclusive",
	"Vogtlaender",
	"AntiParry",
	"AllowTerminalAutoChecks",
	"Checkless",
	"BackHome",

	"WhiteMustCapture",
	"WhiteMustCapture_Ultra",
	"WhiteMustCapture_Anti",
	"BlackMustCapture",
	"BlackMustCapture_Ultra",
	"BlackMustCapture_Anti",
	"WhiteMustCheck",
	"WhiteMustCheck_Ultra",
	"BlackMustCheck",
	"BlackMustCheck_Ultra",
	"WhiteAlphabetic",
	"WhiteAlphabetic_Anti",
	"BlackAlphabetic",
	"BlackAlphabetic_Anti",
	"WhiteMaximummer",
	"WhiteMaximummer_Anti",
	"BlackMaximummer",
	"BlackMaximummer_Anti"
];

// These combinations deserve to be shown in the UI.
var comboNames = [
	"Bichromatic: Monochromatic Monochromatic_Anti",
	"Lortap: Patrol Patrol_Anti",
	"Functionary: Provocation Provocation_Ultra",
	"Unicapture: Multicapture Multicapture_Anti",
	"SuperAndernach: Andernach AntiAndernach",
	"ContraParrain: Circe Circe_Parrain Circe_Mirror",
	"PWC: Circe Circe_Platzwechsel",
	"AllIn: FreeForAll NoImmediateRepetition",
	"ContactGrid: Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"Losing: NonRoyalKings WhiteMustCapture BlackMustCapture",
	"NoWhiteCapture:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra",
	"NoBlackCapture:" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"WhiteMinimummer: WhiteMaximummer WhiteMaximummer_Anti",
	"BlackMinimummer: BlackMaximummer BlackMaximummer_Anti"
];

var otherNames = [
	// Popeye English compatibility.
	"MirrorCirce: Circe Circe_Mirror",
	"Hypervolage: Volage Volage_Hyper",
	"BichromChess: Monochromatic Monochromatic_Anti",
	"MonochromChess: Monochromatic",
	"GridChess: Grid",
	"NoCapture:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"ImmunChess: Immun",
	"ContactGridChess: Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"ChameleonCirce: Circe Circe_Chameleon",
	"CouscousCirce: Circe Circe_Couscous",
	"EquipollentsCirce: Circe Equipollents",
	"FileCirce: Circe Circe_File",
	"DoubleTibet: Tibet Tibet_Double",
	"DiagramCirce: Circe Circe_Diagram",
	"ParrainCirce: Circe Circe_Parrain",
	"SymmetryCirce: Circe Circe_Symmetry",
	"VogtlaenderChess: Vogtlaender",
	"EinsteinChess: Einstein",
	"NewKoeko: Koeko Koeko_New",
	"CirceClone: Circe Circe_Clone",
	"MirrorFileCirce: Circe Circe_File Circe_Mirror",
	"AntipodeanCirce: Circe Circe_Antipodes",
	"MirrorCirceClone: Circe Circe_Clone Circe_Mirror",
	"ReversalEinsteinChess: Einstein Einstein_Reversal",
	"RelegationChess: Relegation",
	"AndernachChess: Andernach",
	"ChameleonChess: Chameleon",
	"FunctionaryChess: Provocation Provocation_Ultra",
	"GlasgowChess: Glasgow",
	"AntiAndernachChess: AntiAndernach",
	"VerticalMirrorCirce: Circe Circe_VerticalMirror",
	"ChecklessChess: Checkless",
	"DiametralCirce: Circe Circe_Diametral",
	"AntiEinsteinChess: Einstein Einstein_Anti",
	"MirrorCouscousCirce: Circe Circe_Couscous Circe_Mirror",
	"CirceAssassin: Circe Circe_Assassin",
	"ExtinctionChess: Extinction",
	"EiffelChess: Eiffel",
	"BlackUltraSchachZwang: BlackMustCheck BlackMustCheck_Ultra",
	"WhiteUltraSchachZwang: WhiteMustCheck WhiteMustCheck_Ultra",
	"AlphabeticChess: WhiteAlphabetic BlackAlphabetic",
	"CirceTurncoats: Circe Circe_Turncoats",
	"CirceDoubleAgents: Circe Circe_Turncoats Circe_Mirror",
	"UltraPatrol: Patrol Patrol_Ultra",
	"AnnanChess: Annan",
	"GenevaChess: Geneva",
	"AntiKoeko: Koeko Koeko_Anti",
	"ProvocationChess: Provocation",
	"WhiteAlphabeticChess: WhiteAlphabetic",
	"BlackAlphabeticChess: BlackAlphabetic",
	"Sentinels_PionAdvers: Sentinels_PionAdverse",
	"Sentinels_ParaSentinelles: Sentinels_Para",
	"Circe_ContraParrain: Circe Circe_Parrain Circe_Mirror",
	"Circe_PWC: Circe_Platzwechsel",

	// Popeye French compatibility.
	"CirceMalefique: Circe Circe_Mirror",
	"Bichromatique: Monochromatic Monochromatic_Anti",
	"Monochromatique: Monochromatic",
	"Grille: Grid",
	"NoirBordCoup: BlackEdgemover",
	"BlancBordCoup: WhiteEdgemover",
	"Patrouille: Patrol",
	"CirceEchange: Circe Circe_Platzwechsel",
	"SansPrises:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"ContactGrille: Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"CirceCameleon: Circe Circe_Chameleon",
	"CirceCouscous: Circe Circe_Couscous",
	"CirceEquipollents: Circe Equipollents",
	"NoirMaximum: BlackMaximummer",
	"NoirMinimum: BlackMaximummer BlackMaximummer_Anti",
	"BlancMaximum: WhiteMaximummer",
	"BlancMinimum: WhiteMaximummer WhiteMaximummer_Anti",
	"Sentinelles: Sentinels",
	"CirceDiagramme: Circe Circe_Diagram",
	"NoirPriseForce: BlackMustCapture",
	"BlancPriseForce: WhiteMustCapture",
	"CirceParrain: Circe Circe_Parrain",
	"CirceSymetrique: Circe Circe_Symmetry",
	"EchecsVogtlaender: Vogtlaender",
	"EchecsEinstein: Einstein",
	"KoekoNouveaux: Koeko Koeko_New",
	"FileCirceMalefique: Circe Circe_File Circe_Mirror",
	"CirceAntipoden: Circe Circe_Antipodes",
	"CirceCloneMalefique: Circe Circe_Clone Circe_Mirror",
	"ReversEchecsEinstein: Einstein Einstein_Reversal",
	"EchecsAndernach: Andernach",
	"EchecsCameleon: Chameleon",
	"EchecsFonctionnaire: Provocation Provocation_Ultra",
	"EchecsGlasgow: Glasgow",
	"EchecsAntiAndernach: AntiAndernach",
	"CirceMalefiqueVerticale: Circe Circe_VerticalMirror",
	"SansEchecs: Checkless",
	"CirceDiametrale: Circe Circe_Diametral",
	"EchecsAntiEinstein: Einstein Einstein_Anti",
	"CirceCouscousMalefique: Circe Circe_Couscous Circe_Mirror",
	"EchecsExtinction: Extinction",
	"EchecsEiffel: Eiffel",
	"NoirUltraSchachZwang: BlackMustCheck BlackMustCheck_Ultra",
	"BlancUltraSchachZwang: WhiteMustCheck WhiteMustCheck_Ultra",
	"BlancSansPrises:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra",
	"NoirSansPrises:" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"EchecsAlphabetiques: WhiteAlphabetic BlackAlphabetic",
	"UltraPatrouille: Patrol Patrol_Ultra",
	"DynastieRoyale: RoyalDynasty",
	"AnnanEchecs: Annan",
	"PionNormale: NormalPawn",	// Typo in Popeye.
	"Elliuortap: Patrol Patrol_Anti",
	"EchecsGeneve: Geneva",
	"QuiPerdGagne: NonRoyalKings WhiteMustCapture BlackMustCapture",
	"EchecsProvacateurs: Provocation",	// Typo in Popeye.
	"RoisKobul: KobulKings",
	"BlancEchecsAlphabetiques: WhiteAlphabetic",
	"NoirEchecsAlphabetiques: BlackAlphabetic",
	"Supergardes: SuperGuards",
	"Retour: BackHome",
	"FaceAFace: FaceToFace",
	"DosADos: BackToBack",
	"Circe_RexInclusif: Circe_RexInclusive",
	"Circe_Malefique: Circe_Mirror",
	"Circe_MalefiqueVerticale: Circe_VerticalMirror",
	"Circe_Colonne: Circe_File",
	"Circe_Symetrie: Circe_Symmetry",
	"Circe_Diagramme: Circe_Diagram",
	"Circe_Echange: Circe_Platzwechsel",
	"Circe_Stricte: Circe_Strict",
	"Madrasi_RexInclusif: Madrasi_RexInclusive",
	"Woozles_RexExclusif: Woozles_RexExclusive",
	"BiWoozles_RexExclusif: BiWoozles_RexExclusive",
	"Bicapture_RexExclusif: Bicapture_RexExclusive",
	"Isardam_RexInclusif: Isardam_RexInclusive",
	"Geneva_RexInclusif: Geneva_RexInclusive",

	// Popeye German compatibility.
	"SpiegelCirce: Circe Circe_Mirror",
	"BichromesSchach: Monochromatic Monochromatic_Anti",
	"MonochromesSchach: Monochromatic",
	"Gitterschach: Grid",
	"SchwarzerRandzueger: BlackEdgemover",
	"WeisserRandzueger: WhiteEdgemover",
	"PatrouilleSchach: Patrol",
	"PlatzwechselCirce: Circe Circe_Platzwechsel",
	"Ohneschlag:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"Immunschach: Immun",
	"KontaktGitter: Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"ChamaeleonCirce: Circe Circe_Chameleon",
	"SchwarzerLaengstzueger: BlackMaximummer",
	"SchwarzerKuerzestzueger: BlackMaximummer BlackMaximummer_Anti",
	"WeisserLaengstzueger: WhiteMaximummer",
	"WeisserKuerzestzueger: WhiteMaximummer WhiteMaximummer_Anti",
	"TibetSchach: Tibet",
	"DoppeltibetSchach: Tibet Tibet_Double",
	"DiagrammCirce: Circe Circe_Diagram",
	"SchwarzerSchlagzwang: BlackMustCapture",
	"WeisserSchlagzwang: WhiteMustCapture",
	"SymmetrieCirce: Circe Circe_Symmetry",
	"VogtlaenderSchach: Vogtlaender",
	"EinsteinSchach: Einstein",
	"NeuKoeko: Koeko Koeko_New",
	"SpiegelFileCirce: Circe Circe_File Circe_Mirror",
	"AntipodenCirce: Circe Circe_Antipodes",
	"SpiegelCirceclone: Circe Circe_Clone Circe_Mirror",
	"ReversesEinsteinSchach: Einstein Einstein_Reversal",
	"Degradierung: Relegation",
	"AndernachSchach: Andernach",
	"ChamaeleonSchach: Chameleon",
	"BeamtenSchach: Provocation Provocation_Ultra",
	"GlasgowSchach: Glasgow",
	"AntiAndernachSchach: AntiAndernach",
	"VertikalesSpiegelCirce: Circe Circe_VerticalMirror",
	"OhneSchach: Checkless",
	"AntiEinsteinSchach: Einstein Einstein_Anti",
	"SpiegelCouscousCirce: Circe Circe_Couscous Circe_Mirror",
	"AssassinCirce: Circe Circe_Assassin",
	"AusrottungsSchach: Extinction",
	"EiffelSchach: Eiffel",
	"SchwarzerUltraSchachZwang: BlackMustCheck BlackMustCheck_Ultra",
	"WeisserUltraSchachZwang: WhiteMustCheck WhiteMustCheck_Ultra",
	"WeisserOhneSchlag:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra",
	"SchwarzerOhneSchlag:" +
		" BlackMustCapture BlackMustCapture_Anti BlackMustCapture_Ultra",
	"AlphabetischesSchach: WhiteAlphabetic BlackAlphabetic",
	"TurncoatCirce: Circe Circe_Turncoats",
	"DoppelAgentenCirce: Circe Circe_Turncoats Circe_Mirror",
	"KoenigsDynastie: RoyalDynasty",
	"Annanschach: Annan",
	"NormalBauern: NormalPawn",
	"GenferSchach: Geneva",
	"Schlagschach: NonRoyalKings WhiteMustCapture BlackMustCapture",
	"ProvokationSchach: Provocation",
	"KontraParrain: Circe Circe_Parrain Circe_Mirror",
	"KobulKoenige: KobulKings",
	"WeissesAlphabetischesSchach: WhiteAlphabetic",
	"SchwarzesAlphabetischesSchach: BlackAlphabetic",
	"SuperDeckungen: SuperGuards",
	"NachHause: BackHome",
	"AngesichtZuAngesicht: FaceToFace",
	"RueckenAnRuecken: BackToBack",
	"Circe_RexInklusive: Circe_RexInclusive",
	"Circe_Spiegel: Circe_Mirror",
	"Circe_VertikalSpiegel: Circe_VerticalMirror",
	"Circe_Chamaeleon: Circe_Chameleon",
	"Circe_Turncoat: Circe_Turncoats",
	"Circe_Symmetrie: Circe_Symmetry",
	"Circe_Diagramm: Circe_Diagram",
	"Circe_Antipoden: Circe_Antipodes",
	"Circe_Strikt: Circe_Strict",
	"Madrasi_RexInklusive: Madrasi_RexInclusive",
	"Woozles_RexExklusive: Woozles_RexExclusive",
	"BiWoozles_RexExklusive: BiWoozles_RexExclusive",
	"Bicapture_RexExklusive: Bicapture_RexExclusive",
	"Isardam_RexInklusive: Isardam_RexInclusive",
	"Geneva_RexInklusive: Geneva_RexInclusive",

	// Popeye applies some of these to the "defending side", but we apply to
	// black as this is the common meaning in publications.
	"Edgemover: BlackEdgemover",
	"Randz\u00FCger: BlackEdgemover",
	"Maximummer: BlackMaximummer",
	"Minimummer: BlackMaximummer BlackMaximummer_Anti",
	"Maximum: BlackMaximummer",	// Keep this otherwise Maximum is ambiguous.
	"Minimum: BlackMaximummer BlackMaximummer_Anti",
	"Laengstzueger: BlackMaximummer",
	"L\u00E4ngstz\u00FCger: BlackMaximummer",
	"K\u00FCrzestz\u00FCger: BlackMaximummer BlackMaximummer_Anti",
	"Schlagzwang: BlackMustCapture",
	"SchachZwang: BlackMustCheck",
	"UltraSchachZwang: BlackMustCheck BlackMustCheck_Ultra",

	// Other English names.
	"ChigorinChess: Chigorin",
	"AlmostChess: Almost",
	"KnightmateChess: Knightmate",
	"NightriderChess: CavalierMajeur",
	"BerolinaChess: Berolina",
	"BerolinaPawns: Berolina",
	"AntiFunctionaryChess: Provocation Provocation_Anti Provocation_Ultra",
	"ReversalChameleon: Chameleon Chameleon_Reversal",
	"SouthernChess: Annan",
	"DoubleEdgemover: WhiteEdgemover BlackEdgemover",
	"MonoWoozles: Woozles",
	"FreeForAllChess: FreeForAll",
	"NoPromotion: PromOnly",
	"MulticaptureChess: Multicapture",
	"UnicaptureChess: Multicapture Multicapture_Anti",
	"AllInChess: FreeForAll NoImmediateRepetition",
	"InterchangeCirce: Circe Circe_Platzwechsel",
	"AntiNewKoeko: Koeko Koeko_Anti Koeko_New",
	"K\u00F6ko: Koeko",
	"AntiK\u00F6ko: Koeko Koeko_Anti",
	"NewK\u00F6ko: Koeko Koeko_New",
	"AntiNewK\u00F6ko: Koeko Koeko_Anti Koeko_New",
//	"MustCapture: WhiteMustCapture BlackMustCapture",
	"LosingChess: NonRoyalKings WhiteMustCapture BlackMustCapture",
//	"MustCheck: WhiteMustCheck BlackMustCheck",
	"WhiteUltraMustCheck: WhiteMustCheck WhiteMustCheck_Ultra",
	"BlackUltraMustCheck: BlackMustCheck BlackMustCheck_Ultra",
//	"UltraMustCheck:" +
//		" WhiteMustCheck WhiteMustCheck_Ultra" +
//		" BlackMustCheck BlackMustCheck_Ultra",
	"AlphabeticalChess: WhiteAlphabetic BlackAlphabetic",
	"AntiAlphabeticChess:" +
		" WhiteAlphabetic WhiteAlphabetic_Anti" +
		" BlackAlphabetic BlackAlphabetic_Anti",
	"AntiAlphabeticalChess:" +
		" WhiteAlphabetic WhiteAlphabetic_Anti" +
		" BlackAlphabetic BlackAlphabetic_Anti",
	"DoubleMaximummer: WhiteMaximummer BlackMaximummer",
	"DoubleMinimummer:" +
		" WhiteMaximummer WhiteMaximummer_Anti" +
		" BlackMaximummer BlackMaximummer_Anti",
	"Circe_Exchange: Circe_Platzwechsel",
	"Circe_Interchange: Circe_Platzwechsel",

	// Other French names.
	"Echecs960: Chess960",
	"B\u00E9rolina: Berolina",
	"PionsB\u00E9rolina: Berolina",
	"CylindreVertical: VerticalCylinder",
	"Provocateurs: Provocation",
	"Fonctionnaires: Provocation Provocation_Ultra",
	"AntiFonctionnaires: Provocation Provocation_Anti Provocation_Ultra",
	"Bicaptures: Bicapture",
	"Cannibales: Cannibal",
	"SansPromotion: PromOnly",
	"PromotionsAdverses: AdversePromotions",
	"Multicaptures: Multicapture",
	"Unicaptures: Multicapture Multicapture_Anti",
	"Cam\u00E9l\u00E9on: Chameleon",
	"D\u00E9gradation: Relegation",
	"Circ\u00E9: Circe",
	"Circ\u00E9Mal\u00E9fique: Circe Circe_Mirror",
	"Circ\u00E9Cam\u00E9l\u00E9on: Circe Circe_Chameleon",
	"Alphab\u00E9tiques: WhiteAlphabetic BlackAlphabetic",
	"AntiAlphab\u00E9tiques:" +
		" WhiteAlphabetic WhiteAlphabetic_Anti" +
		" BlackAlphabetic BlackAlphabetic_Anti",
	"MaximumBlanc: WhiteMaximummer",
	"Circe_Mal\u00E9fique: Circe_Mirror",
	"Circe_Cam\u00E9l\u00E9on: Circe_Chameleon",

	// Other German names.
	"Schach960: Chess960",
	"Berolinaschach: Berolina",
	"K\u00F6nigsDynastie: RoyalDynasty",
	"Wei\u00DFerRandz\u00FCger: WhiteEdgemover",
	"SchwarzerRandz\u00FCger: BlackEdgemover",
	"DoppelRandz\u00FCger: WhiteEdgemover BlackEdgemover",
	"AntiBeamtenschach: Provocation Provocation_Anti Provocation_Ultra",
	"Kannibalenschach: Cannibal",
	"OhneUW: PromOnly",
	"Cham\u00E4leonschach: Chameleon",
	"Kobulk\u00F6nige: KobulKings",
	"Cham\u00E4leonCirce: Circe Circe_Chameleon",
	"Neuk\u00F6ko: Koeko Koeko_New",
	"Antineuk\u00F6ko: Koeko Koeko_Anti Koeko_New",
	"Giko: Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"Gitterkontaktschach:" +
		" Grid Koeko Koeko_AffectedByGrid Koeko_AppliesToRebirth",
	"Vogtl\u00E4nderSchach: Vogtlaender",
	"Wei\u00DFerOhneSchlag:" +
		" WhiteMustCapture WhiteMustCapture_Anti WhiteMustCapture_Ultra",
	"Alphabetschach: WhiteAlphabetic BlackAlphabetic",
	"Wei\u00DFerL\u00E4ngstz\u00FCger: WhiteMaximummer",
	"Wei\u00DFerK\u00FCrzestz\u00FCger: WhiteMaximummer WhiteMaximummer_Anti",
	"Doppell\u00E4ngstz\u00FCger: WhiteMaximummer BlackMaximummer",
	"Doppelk\u00FCrzestz\u00FCger:" +
		" WhiteMaximummer WhiteMaximummer_Anti" +
		" BlackMaximummer BlackMaximummer_Anti",
	"Circe_Cham\u00E4leon: Circe_Chameleon"
];

var DEBUG = false;	// To turn on some console.log calls.

// Helper functions.

function isWhiteSpace(c) {
	return (c == " " || c == "\t" || c == "\n" || c == "\r");
}

// Searches "token" in "list", returning value, or one of these special values:
// "unknown", "ambiguous".
function findInList(token, list) {
	var matchTokens = [];
	var matchValue = "unknown";
	var i;
	for (i = 0; i < list.length; i++) {
		var parts = list[i].split(": ");
		var listToken = parts[0];
		var listValue = parts[parts.length - 1];

		if (token.length > listToken.length)
			continue;
		if (token.toLowerCase() ==
		    listToken.substr(0, token.length).toLowerCase()) {
			if (matchTokens.length == 0) {
				matchTokens.push(listToken);
				matchValue = listValue;
			}
			else if (matchValue == listValue) {
				// It gives the same answer anyway.
				matchTokens.push(listToken);
			}
			else {
				var j;
				for (j = 0; j < matchTokens.length; j++) {
					if (matchTokens[j] ==
					    listToken.substr(0, matchTokens[j].length)) {
						// If the new match contains an old match, ignore.
						// E.g. Circ will match Circe first, next CirceClone
						// which we ignore.
						break;
					}
				}
				if (j == matchTokens.length) {
					if (DEBUG)
						console.log(listToken + " " + matchTokens);
					return "ambiguous";
				}
			}
		}
	}
//	console.log(token + " in " + list + " found " + matchValue);
	return matchValue;
}

// Returns null if there is no match.
function grabLongestMatch(token, list) {
	var match = null;
	for (var i = 0; i < list.length; i++) {
		var parts = list[i].split(": ");
		if (token.substr(0, parts[0].length).toLowerCase() == parts[0] &&
		    (match == null || parts[0].length > match.length)) {
			match = parts[0];
		}
	}
	return match;
}

function grabClosure(token) {
	var i = 0;
	var c = token.charAt(i).toLowerCase();

	while (true) {
		if (c == "h" || c == "*") {
			i++;
			c = token.charAt(i);
		}
		else if (c == "s" || c == "r") {
			i++;
			break;
		}
		else
			break;
	}

	return token.substr(0, i);
}

// "name" is only used for the error message.
function readType(parser, command, name, list) {
	var token = parser.getCurrToken();
	if (token == "") {		// Undefined or empty.
		alert("Error: " + name + " type is missing.");
		return false;
	}
	var separator = parser.getCurrSeparator();

	var type = findInList(token, list);
	if (type == "unknown" || type == "ambiguous") {
		alert("Error: " + name + " type '" + token + "' is " + type + ".");
		return false;
	}

	command.type = {text: separator + token,
					value: type};
	parser.curr++;
	return true;
}

// "name" is only used for the error message.
function readAnyOptions(parser, command, name, list) {
	command.options = [];

	while (true) {
		var token = parser.getCurrToken();
		if (token == "")		// Undefined or empty.
			break;
		var separator = parser.getCurrSeparator();

		var option = findInList(token, list);
		if (option == "ambiguous") {
			alert("Error: " + name + " option '" + token + "' is " +
			      option + ".");
			return false;
		}
		else if (option == "unknown")
			break;

		command.options.push({text: separator + token,
			                  value: option});
		parser.curr++;	
	}
	return true;
}

// Parser functions.

function parseComment(parser, command) {
	while (parser.curr < parser.parts.length) {
		var separator = parser.getCurrSeparator();
		if (separator.search("[\n\r]") >= 0)
			break;
		var token = parser.getCurrToken();
		command.text += separator + token;
		parser.curr++;
	}
}

function parseForsyth(parser, command) {
	// Parse board.
	var token = parser.getCurrToken();
	if (token == "") {		// Undefined or empty.
		alert("Error: forsyth string is missing.");
		return false;
	}
	var separator = parser.getCurrSeparator();

	var T = pieceLetters[parser.language];
	var EMPTY = 0;
	var WHITE = 32;
	var BLACK = 64;

	var board = [];
	var wRoyal = -1;
	var bRoyal = -1;
	for (var i = 0; i < token.length; i++) {
		// Grab a number if there is one.
		var n = 0;
		while (true) {
			var d = token.charCodeAt(i);
			if (d >= 48 && d < 58) {
				n = n * 10 + (d - 48);
				i++;
				if (i < token.length)
					continue;
			}
			break;
		}
		if (board.length + n > 64) {
			alert("Error: forsyth number '" + n + "' exceeds 64 squares.");
			return false;
		}
		while (n > 0) {
			board.push(EMPTY);
			n--;
		}

		if (i == token.length)
			break;

		var c = token.charAt(i);
		if (c == "/") {
			if (board.length == 0 || board.length == 64 ||
			    board.length % 8 != 0) {
				alert("Error: forsyth row separator '/' is badly placed.");
				return false;
			}
			continue;
		}

		// Grab a piece letter.
		var j;
		for (j = 0; j < T.length; j++) {
			if (c == T[j]) {
				board.push(WHITE + j);
				break;
			}
			else if (c == T[j].toLowerCase()) {
				board.push(BLACK + j);
				break;
			}
		}
		if (j == T.length) {
			alert("Error: forsyth piece letter '" + c + "' is unknown.");
			return false;
		}
		if (board.length > 64) {
			alert("Error: forsyth piece letter '" + c +
			      "' exceeds 64 squares.");
			return false;
		}
	}

	if (board.length < 64) {
		alert("Error: forsyth string fills only " + board.length + " squares.");
		return false;
	}

	command.position = {text: separator + token,
					    value: {board: board,
					            wRoyal: wRoyal,
					            bRoyal: bRoyal}};
	parser.curr++;

	return true;
}

// Process a piece-and-squares token such as "Sb1g1". Returns false if the
// token isn't of that form.
function ParsePieceAndSquares(token, language, color, position) {
	var WHITE = 32;
	var BLACK = 64;

	if (token.length < 3)
		return false;

	var pieceLetter = token.substr(0, (token.length + 1) % 2 + 1).toUpperCase();

	var piece = pieceLetters[language].indexOf(pieceLetter);
	if (piece < 0)
		return false;

	var royal = false;	// TODO: this will become a parameter.
	// Note: with the current design we won't be able to set royal pieces
	// with ColorThePieces.

	if (color == "white")
		piece += WHITE;
	else if (color == "black")
		piece += BLACK;
	else {
		alert("Error: unknown color '" + color + "'");
		return false;
	}

	var squares = [];
	for (var i = pieceLetter.length; i < token.length; i += 2) {
		var file = token.charCodeAt(i);
		var rank = token.charCodeAt(i+1);
		if (file == 231) file = 99;	// c
		if (file == 233) file = 101;	// e
		if (file < 96) file += 32;	// toLowerCase
		file -= 97;	// a
		rank -= 49;	// 1
		if (file >= 0 && file < 8 && rank >= 0 && rank < 8)
			squares.push((7-rank) * 8 + file);
		else
			return false;
	}

	// Everything is good, add pieces to the board.
	for (var i = 0; i < squares.length; i++) {
		position.board[squares[i]] = piece;
		if (royal) {
			if (color == "white")
				position.wRoyal = squares[i];
			else if (color == "black")
				position.bRoyal = squares[i];
		}
	}

	return true;
}

function parsePieces(parser, command) {
	var position = {
		board: [],
		wRoyal: -1,
		bRoyal: -1
	};

	var EMPTY = 0;
	for (var i = 0; i < 64; i++)
		position.board[i] = EMPTY;

	var text = "";

	var token = parser.getCurrToken();

	while (true) {
		if (token == "")
			break;	// Maybe we're done.
		var separator = parser.getCurrSeparator();

		var color = findInList(token, pieceColorNames[parser.language]);
		if (color == "ambiguous") {
			alert("Error: color name '" + token + "' is " + color + ".");
			return false;
		}
		else if (color == "unknown")
			break;	// Maybe we're done and token is the next command.

		// Color token is accepted!
		text += separator + token;

		// Parse pieces until the token isn't a piece anymore.
		while (true) {
			parser.curr++;

			token = parser.getCurrToken();
			if (!ParsePieceAndSquares(token, parser.language, color, position))
				break;
			separator = parser.getCurrSeparator();

			text += separator + token;
		}
	}

	command.position = {text: text, value: position};
	return true;
}

function parseOption(parser, command) {
	if (!readType(parser, command, "option", optionNames))
		return false;

	var type = command.type.value;
	if (type == "FirstToMove") {
		// Parse color.
		var token = parser.getCurrToken();
		if (token == "") {		// Undefined or empty.
			alert("Error: FirstToMove color is missing.");
			return false;
		}
		var separator = parser.getCurrSeparator();

		var color = findInList(token, pieceColorNames[parser.language]);
		if (color == "unknown" || color == "ambiguous") {
			alert("Error: color name '" + token + "' is " + color + ".");
			return false;
		}

		command.color = {text: separator + token,
						 value: color};
		parser.curr++;
	}
	else if (type == "InitialMoveNumber") {
		// Parse number.
		var token = parser.getCurrToken();
		if (token == "") {		// Undefined or empty.
			alert("Error: InitialMoveNumber value is missing.");
			return false;
		}
		var separator = parser.getCurrSeparator();

		var re = /^-?\d*(|\.0|\.5)$/;
		if (!re.test(token)) {
			alert("Error: InitialMoveNumber '" + token + "' is wrong.");
			return false;
		}

		var number = parseFloat(token);

		command.number = {text: separator + token,
						  value: number};
		parser.curr++;
	}

	return true;
}

function parsePlay(parser, command) {
	if (!readType(parser, command, "play", playNames))
		return false;

	command.type.value = parseInt(command.type.value);

	// Parse length.
	var token = parser.getCurrToken();
	if (token == "") {		// Undefined or empty.
		alert("Error: play length is missing.");
		return false;
	}
	var separator = parser.getCurrSeparator();

	var re = /^\d*(|\.0|\.5)$/;
	if (!re.test(token)) {
		alert("Error: play length '" + token + "' is wrong.");
		return false;
	}
	parser.curr++;

	var length = parseFloat(token) * 2;

	// If a series, then reduce integer lengths by 1 ply.
	var series = (command.type.value ==
	              parseInt(findInList("series", playNames)));
	if (series) {
		re = /^\d*$/;
		if (re.test(token))
			length -= 1;
	}

	command.length = {text: separator + token,
	                  value: length};

	return true;
}

// "name" is typically "dia", "a=>b", or "PG" to make the error message less
// confusing.
function parseDia(parser, command, name) {
	// We require a diagram.
	var token = parser.getCurrToken();
	if (token == "") {		// Undefined or empty.
		alert("Error: " + name + " diagram is missing.");
		return false;
	}
	var separator = parser.getCurrSeparator();

	var type = findInList(token, diagramNames);
	command.dia = {text: separator + token,
		           command: type};
	parser.curr++;

	if (type == "forsyth") {
		if (!parseForsyth(parser, command.dia))
			return false;
	}
	else if (type == "pieces") {
		if (!parsePieces(parser, command.dia))
			return false;
	}
	else {
		alert("Error: " + name +
		      " must be immediately followed by forsyth or pieces.");
		return false;
	}

	return readAnyOptions(parser, command.dia, name, diagramOptions);
}

function parseTest(parser, command) {
	// Multiple passes in case there is a not.
	var aim;
	for (var i = 0; ; i++) {
		// Parse aim.
		var token = parser.getCurrToken();
		if (token == "") {		// Undefined or empty.
			alert("Error: aim is missing.");
			return false;
		}
		var separator = parser.getCurrSeparator();

		aim = findInList(token, testNames);

		if (i == 0 && aim == "not") {
			command.negated = {text: separator + token,
		                	   value: true};
			parser.curr++;
		}
		else
			break;
	}

	if (aim == "unknown" || aim == "ambiguous" || aim == "not") {
		alert("Error: aim '" + token + "' is " + aim + " .");
		return false;
	}

	command.aim = {text: separator + token,
				   value: parseInt(aim)};
	parser.curr++;

	if (aim == "0")	// dia.
		return parseDia(parser, command, token);

	return true;
}

function parseSet(parser, command) {
	if (!readType(parser, command, "set", setNames))
		return false;

	command.type.value = parseInt(command.type.value);

	// Set is just a test with a type.
	if (!parseTest(parser, command))
		return false;

	if (command.type.value == parseInt(findInList("pass", setNames)) &&
	    (command.aim.value == parseInt(findInList("x", testNames)) ||
	     command.aim.value == parseInt(findInList("%", testNames)) ||
	     command.aim.value == parseInt(findInList("00", testNames)) ||
	     command.aim.value == parseInt(findInList("ep", testNames)))) {
		alert("Error: 'set pass' doesn't currently work with the aims " +
		      "'x', '%', '00', and 'ep'.");
		return false;
	}

	return readAnyOptions(parser, command, "set", setOptions);
}

// Helper variable and functions to build subcommands in parseStipulation().

var commandBlackFirst = {command: "option",
                         text: "\noption",
                         type: {text: " FirstToMove", value: "FirstToMove"},
                         color: {text: " black", value: "black"}};

var commandNotationGap = {command: "option",
                          text: "\noption",
                          type: {text: " InitialMoveNumber",
                                 value: "InitialMoveNumber"},
                          number: {value: 1.5}};

// Length of series have already been subtracted by 0.5.
function commandPlay(type, length) {
	var plies = length * 2;

	var lengthStr = (type == "series" && length % 1 == 0.5) ?
	                (length + 0.5).toFixed(0) :
					length.toFixed(1);

	var id = parseInt(findInList(type, playNames));

	var ret = {command: "play",
	           text: "\nplay",
	           type: {text: " " + type,
	                  value: id},
	           length: {text: " " + lengthStr,
	                    value: plies}};
	return ret;
}

// Copy test from command.
function commandTest(command) {
	var ret = {command: "test",
	           text: "\ntest",
			   aim: {text: " " + command.aim.text,
			         value: command.aim.value}};

	if ("negated" in command)
		ret.negated = command.negated;

	if (command.aim.value == 0)
		ret.dia = command.dia;

	return ret;
}

// Copy test from command.
function commandSetTestReflex(command) {
	var ret = commandTest(command);
	ret.command = "set";
	ret.text = "\nset";
	ret.type = {text: " test",
	            value: parseInt(findInList("test", setNames))};
	ret.options = [{text: " reflex", value: findInList("reflex", setOptions)}];
	return ret;
}

function commandSetPass(negated, aim) {
	var ret = {command: "set",
	           text: "\nset",
	           type: {text: " pass",
			          value: parseInt(findInList("pass", setNames))},
	           aim: {text: " " + aim,
	                 value: parseInt(findInList(aim, testNames))},
	           options: []};
	if (negated) {
		ret.negated = {text: " not",
		               value: true};
	}
	return ret;
}

function addClosureSubCommands(subcommands, closure) {
	for (var i = 0; i < closure.length; i++) {
		var c = closure.charAt(i);
		if (c == "h")
			subcommands.push(commandPlay("help", 0.5));
		else if (c == "*")
			subcommands.push(commandPlay("pass", 0.5));
		else if (c == "s" || c == "r")
			subcommands.push(commandPlay("defensive", 0.5));
		else
			return false;
	}

	return true;
}

function parseStipulation(parser, command) {
	var token = parser.getCurrToken();
	if (token == "") {		// Undefined or empty.
		alert("Error: stipulation type is missing.");
		return false;
	}
	var separator = parser.getCurrSeparator();
	parser.curr++;

	// Identify the stipulation as prefix + closure + aimName.
	// Store the rest in lengthStr in case the length is part of the token.

	// prefix != null because the empty string is one of the options.
	var prefix = grabLongestMatch(token, prefixNames);
	var closure = grabClosure(token.substr(prefix.length));
	var aimName = grabLongestMatch(token.substr(prefix.length + closure.length),
	                               testNames);

	if (aimName == null) {
		alert("Error: stipulation '" + token + "' is unknown.");
		return false;
	}
	var aim = parseInt(findInList(aimName, testNames));
	var lengthStr = token.substr(prefix.length + closure.length +
	                             aimName.length);

	var separator2 = "";
	if (lengthStr == "") {
		// Grab next token.
		var token2 = parser.getCurrToken();
		if (token2 == "") {		// Undefined or empty.
			alert("Error: stipulation length is missing.");
			return false;
		}
		var separator2 = parser.getCurrSeparator();
		parser.curr++;

		lengthStr = token2;
	}

	// Store aim now, because it is copied in subcommand in different ways.
	command.prefix = {text: separator + prefix,
	                  value: prefix};	// Value used by addAntiParryIfNeeded.

	command.closure = {text: closure};	// Value unused.

	command.aim = {text: aimName,
				   value: aim};

	command.length = {text: separator2 + lengthStr};

	if (aim == 0) {
		if (!parseDia(parser, command, aimName))
			return false;
	}

	var isHelp = (prefix == "h" || prefix == "ph" || prefix == "ah" ||
	              prefix == "pah");
	var isSeries = (prefix == "ser-" || prefix == "phser-" ||
	                prefix == "ahser-" || prefix == "pahser-");

	// Parse length.
	var re = /^\d*(|\.0|\.5)$/;
	if (!re.test(lengthStr)) {
		alert("Error: stipulation length '" + lengthStr + "' is wrong.");
		return false;
	}

	var length = parseFloat(lengthStr);
	if (isSeries) {
		re = /^\d*$/;
		var hasDecimals = !re.test(lengthStr);
		if (!hasDecimals)
			length -= 0.5;
	}

	var hasHalfMove = (length % 1 == 0.5);

	// Translate the compact conventional stipulations into elementary
	// commands.
	var subcommands = [];

	// Has pass moves.
	if (prefix == "phser-" || prefix == "pahser-" ||
	    prefix == "ph" || prefix == "pah") {
		subcommands.push(commandSetPass(true, "+"));
	}
	if (prefix == "ahser-" || prefix == "pahser-" ||
	    prefix == "ah" || prefix == "pah") {
		subcommands.push(commandSetPass(true, "!+"));
	}
	if (prefix == "ser-" ||
	    (subcommands.length == 0 && closure.indexOf("*") >= 0)) {
		subcommands.push(commandSetPass(false, "~"));
	}

	// Closure information.
	var lastHelpIndex = closure.lastIndexOf("h");
	if (lastHelpIndex == -1 && isSeries && !hasHalfMove)
		lastHelpIndex--;	// Series prefix is h*h*h*.
	var lastMove = closure.charAt(closure.length - 1);
	var lastMoveBlack = (lastMove == "s" || lastMove == "r" ||
	                     (closure.length - lastHelpIndex) % 2 == 0);

	// Add prefix subcommands.
	if (prefix == "") {
		if (closure.length != 0) {
			alert("Closure must be empty in direct or PG genres.");
			return false;
		}

		if (aim == 0) {
			// Empty-dia is treated as a help game without
			// any blackFirst convention for integer length.
			if (length > 0)
				subcommands.push(commandPlay("help", length));
		}
		else {
			// For everything else, Empty-X is direct-X.
			if (length != 1) {
				alert("Error: direct stipulation length must be exactly " +
				      "1 in Jacobi");
				return false;
			}
			subcommands.push(commandPlay("help", 0.5));	
		}
		subcommands.push(commandTest(command));
	}
	else if (prefix == "s" || prefix == "r") {
		if (closure.length != 0) {
			alert("Closure must be empty in 's' or 'r' genres.");
			return false;
		}
		if (length != 1) {
			alert("Error: " + prefix +
			      " length must be exactly 1 in Jacobi");
			return false;
		}
		if (prefix == "r")
			subcommands.push(commandSetTestReflex(command));

		subcommands.push(commandPlay("help", 0.5));
		subcommands.push(commandPlay("defense", 0.5));
		subcommands.push(commandTest(command));
	}
	else if (isHelp) {
		var prefixLength = length - closure.length * 0.5;

		if (hasHalfMove)
			subcommands.push(commandNotationGap);
//		alert("h = " + hasHalfMove + ", l = " + lastMoveBlack + ", c " + closure.length + " , l = " + lastHelpIndex);
		if (hasHalfMove == lastMoveBlack)
			subcommands.push(commandBlackFirst);

		if (prefixLength < 0) {
			alert("Error: " + token + " length must be longer");
			return false;
		}

		if (lastMove == "r")
			subcommands.push(commandSetTestReflex(command));

		if (prefixLength > 0)
			subcommands.push(commandPlay("help", prefixLength));
		if (!addClosureSubCommands(subcommands, closure)) {
			alert("Problem with closure '" + closure + "'");
			return false;
		}
		subcommands.push(commandTest(command));

/*
		else
			subcommands.push(commandBlackFirst);
		if (length > 0)
			subcommands.push(commandPlay("help", length, []));
		subcommands.push(commandTest(command));
//		}
//		else if (genre == "hs" || genre == "hr") {
		if (length < 0.5) {
			alert("Error: " + genre + " length must be at least 0.5");
			return false;
		}
		if (hasHalfMove) {
			subcommands.push(commandBlackFirst);
			subcommands.push(commandNotationGap);
		}
		if (genre == "hs") {
			if (length > 0.5)
				subcommands.push(commandPlay("help", length - 0.5, []));
			subcommands.push(commandPlay("defense", 0.5, []));
			subcommands.push(commandTest(command));
		}
		else {
			subcommands.push(commandSetTest(command));
			if (length > 0.5) {
				subcommands.push(commandPlay("help", length - 0.5,
					                         ["none"]));
			}
			subcommands.push(commandPlay("help", 0.5, ["accept"]));	
		}
		*/
	}

	else if (isSeries) {
		var prefixLength = length;
		if (prefixLength < 0.5) {
			alert("Error: " + token + " length must be longer");
			return false;
		}

		if ((lastMoveBlack && closure.length % 2 == 0) ||
		    (!lastMoveBlack && closure.length % 2 != 0)) {
			subcommands.push(commandBlackFirst);
		}

		if (lastMove == "r")
			subcommands.push(commandSetTestReflex(command));

		subcommands.push(commandPlay("series", prefixLength));
		if (!addClosureSubCommands(subcommands, closure)) {
			alert("Problem with closure '" + closure + "'");
			return false;		
		}
		subcommands.push(commandTest(command));

/*
	// An initial star is equivalent to increasing the length by 0.5.
//		if (genre == "*s" || genre == "*r") {
		if (!hasHalfMove) {
			alert("Error: " + series + genre + " length can't be " +
				  lengthStr);
			return false;
		}
		length += 0.5;
		hasHalfMove = false;
		genre = genre.substr(1);
//		}

//		if (genre == "") {
		if (length < 0) {	// User probably typed "0" which became -0.5.
			alert("Error: " + series + genre + " length can't be " +
				  lengthStr);
			return false;
		}
		subcommands.push(commandPlay("help", length, ["series"]));
		subcommands.push(commandTest(command));
//		}
//		else if (genre == "s" || genre == "r") {
		if (length < 0) {
			alert("Error: " + series + genre + " length can't be " +
				  lengthStr);
			return false;
		}
		if (!hasHalfMove)	// length n.0
			subcommands.push(commandBlackFirst);
		if (genre == "s") {
			subcommands.push(commandPlay("help", length, ["series"]));
			subcommands.push(commandPlay("defense", 0.5, []));
			subcommands.push(commandTest(command));
		}
		else {
			subcommands.push(commandSetTest(command));
			subcommands.push(commandPlay("help", length,
				                         ["none", "series", "wake"]));
			subcommands.push(commandPlay("help", 0.5, ["accept"]));
		}
//		}
//		else if (genre == "h") {
		// It's stupid to call this with a fraction because n.0 is the
		// same as a slightly longer ser-, but we don't judge.
		if (length < 0) {	// User probably typed "0" which became -0.5.
			alert("Error: " + series + genre +
			      " length must be a position integer");
			return false;
		}
		if (hasHalfMove)	// length n or n.5
			subcommands.push(commandBlackFirst);
		subcommands.push(commandPlay("help", length, ["series"]));
		subcommands.push(commandPlay("help", 0.5, []));
		subcommands.push(commandTest(command));
//		}
//		else if (genre == "hs" || genre == "hr") {
		// It's stupid to call this with a fraction because n.0 is the
		// same as a slightly longer ser-h, but we don't judge.
		if (length < 0) {
			alert("Error: " + series + genre + " length can't be " +
				  lengthStr);
			return false;
		}

		if (hasHalfMove)	// length n or n.5
			subcommands.push(commandBlackFirst);

		// This follows Popeye. Magazines sometimes do length-1 here, which
		// is consistent with length-0.5 in hs, but inconsistent with
		// ser-h and ser-s where it should then be length-0.5.
		if (genre == "hs") {
			subcommands.push(commandPlay("help", length, ["series"]));
			subcommands.push(commandPlay("help", 0.5, []));
			subcommands.push(commandPlay("defense", 0.5, []));
			subcommands.push(commandTest(command));
		}
		else {
			subcommands.push(commandSetTest(command));
			subcommands.push(commandPlay("help", length,
				                         ["none", "series", "wake"]));
			subcommands.push(commandPlay("help", 0.5, ["none"]));
			subcommands.push(commandPlay("help", 0.5, ["accept"]));	
		}
		*/
	}
	else {
		alert("Unknown stipulation prefix '" + prefix + "'");
		return false;
	}

	command.subcommands = subcommands;

	return true;
}

function addConditionToList(list, cond, prefix) {
	var parts = cond.split(": ");
	var parts2 = parts[0].split("_");

	if (prefix == "") {
		// Only accept conditions without prefix.
		if (parts2.length == 1)
			list.push(cond);
	}
	else {
		// Only accept conditions with matching prefix.
		if (parts2.length == 2 && prefix == parts2[0])
			list.push(parts2[1] + ": " + (parts.length == 1 ? cond : parts[1]));
	}
}

function conditionBase(name) {
	var i = name.indexOf("_");
	return (i >= 0) ? name.substr(0, i) : name;
}

function conditionTail(name) {
	var i = name.indexOf("_");
	return (i >= 0) ? name.substr(i + 1) : name;
}

// Condition is a space separated list. Returns an array of distinct prefixes.
function addConditionBases(set, cond) {
	var parts = cond.split(" ");
	for (var i = 0; i < parts.length; i++) {
		var prefix = conditionBase(parts[i]);
		if (set.indexOf(prefix) == -1)
			set.push(prefix);
	}
}

function parseCondition(parser, command) {
	command.conditions = [];

	var prefixSet = [];

	while (true) {
		var token = parser.getCurrToken();
		if (token == "")
			break;	// Ignore undefined or whitespace after last keyword.
		var separator = parser.getCurrSeparator();

		var cond = "unknown";

		// First priority: prefixed conditions.
		// Only allow a suffix after a prefix set of size 1, otherwise
		// if we allow "MustCheck Ultra", then it's too hard to break it
		// apart because the components of two tokens must be reorganized.
		if (prefixSet.length == 1) {
			var list = [];
			for (var i = 0; i < conditionNames.length; i++)
				addConditionToList(list, conditionNames[i], prefixSet[0]);
			for (var i = 0; i < comboNames.length; i++)
				addConditionToList(list, comboNames[i], prefixSet[0]);
			for (var i = 0; i < otherNames.length; i++)
				addConditionToList(list, otherNames[i], prefixSet[0]);

			cond = findInList(token, list);
			if (cond == "ambiguous") {
				alert("Error: condition '" + token + "' is " + cond + ".");
				return false;
			}
		}

		// Second priority: non-prefixed conditions.
		if (cond == "unknown") {
			var list = [];
			for (var i = 0; i < conditionNames.length; i++)
				addConditionToList(list, conditionNames[i], "");
			for (var i = 0; i < comboNames.length; i++)
				addConditionToList(list, comboNames[i], "");
			for (var i = 0; i < otherNames.length; i++)
				addConditionToList(list, otherNames[i], "");

			cond = findInList(token, list);

			if (cond == "ambiguous") {
				alert("Error: condition '" + token + "' is " + cond + ".");
				return false;
			}
		}

		if (cond == "unknown")
			break;	// Current token is not a cond.

		command.conditions.push({text: separator + token,
								 value: cond});
		parser.curr++;

		// Update prefix set.
		prefixSet = [];
		addConditionBases(prefixSet, cond);
	}

	return true;
}

// Returns true on success.
function parseCommand(parser, commands) {
	var separator = parser.getCurrSeparator();
	var token = parser.getCurrToken();

	var command;
	if (token == "")	// Empty whitespace at the end.
		command = "ignore";
	else {
		// From token to command.
		command = findInList(token, commandNames);
		if (command == "unknown" || command == "ambiguous") {
			alert("Error: command '" + token + "' is " + command + ".");
			return false;
		}
	}

	var commandObj = {text: separator + token,
					  command: command};
	commands.push(commandObj);
	parser.curr++;

	if (command == "ignore")
		return true;
	if (command == "englishS" || command == "englishN" ||
	    command == "french" || command == "german") {
		parser.language = command;

		// Turn it into a "language" command so we don't have so many
		// different commands.
		commandObj.command = "language";
		commandObj.type = command;
		return true;
	}
	else if (command == "comment") {
		parseComment(parser, commandObj);
		commandObj.command = "ignore";	// Solver doesn't need to know.
		return true;
	}
	else if (command == "forsyth")
		return parseForsyth(parser, commandObj);
	else if (command == "pieces")
		return parsePieces(parser, commandObj);
	else if (command == "option")
		return parseOption(parser, commandObj);
	else if (command == "play")
		return parsePlay(parser, commandObj);
	else if (command == "test")
		return parseTest(parser, commandObj);
	else if (command == "set")
		return parseSet(parser, commandObj);
	else if (command == "stipulation")
		return parseStipulation(parser, commandObj);
	else if (command == "condition")
		return parseCondition(parser, commandObj);
	else {
		alert("Error: command '" + token + "' is " + command +
		      " but has no parser.");
		return false;
	}
}

function textFromCommands(commands) {
	var ret = "";
	for (var i = 0; i < commands.length; i++) {
		var command = commands[i];
		ret += command.text;
		if (command.command == "forsyth" || command.command == "pieces")
			ret += command.position.text;
		else if (command.command == "option") {
			ret += command.type.text;
			if (command.type.value == "FirstToMove")
				ret += command.color.text;
			else if (command.type.value == "InitialMoveNumber")
				ret += command.number.text;
		}
		else if (command.command == "play") {
			ret += command.type.text + command.length.text;
		}
		else if (command.command == "test" ||
		         command.command == "set" ||
		         command.command == "stipulation") {
			if (command.command == "set")
				ret += command.type.text;

			if ("negated" in command)
				ret += command.negated.text;

			if (command.command == "stipulation")
				ret += command.prefix.text + command.closure.text;

			ret += command.aim.text;

			if (command.command == "stipulation")
				ret += command.length.text;

			if (command.aim.value == 0) {	// dia
				ret += command.dia.text + command.dia.position.text;
				for (var j = 0; j < command.dia.options.length; j++)
					ret += command.dia.options[j].text;
			}

			if (command.command == "set") {
				for (var j = 0; j < command.options.length; j++)
					ret += command.options[j].text;
			}
		}
		else if (command.command == "condition") {
			for (var j = 0; j < command.conditions.length; j++)
				ret += command.conditions[j].text;
		}
	}
	return ret;
}

// ---------------------------------------------------------------------------

// Functions from engine. These are undefined for now, and will get set
// asynchronously. They will be available when g_engineReady = true.

var SetProblemFromJSON;
var GetLastPly;
var SP_PlayMove;
var SP_ReverseMove;
var SP_GetLastMoveNotation;
var Solver_PawnOnIllegalRank;
var Solver_CanCaptureTheKing;
var Solver_PrepareDemolitionPass;
var Solver_IsPassIncomplete;
var Solver_RunStep;
var Solver_SearchFraction;
var Solver_GetRow960;
var TT_LoadFraction = function() {return 0};	// Initialized for v0.4.3.

// ---------------------------------------------------------------------------

var g_hashMem;
var g_useWasm;
var g_engineLaunched = false;
var g_engineReady = false;

var g_commands = [];
var g_language = "englishS";
var g_firstPly = 0;
var g_lastPly = 0;
var g_initialMoveNumber = 1;

var g_paused = false;
var g_lastCurrTime = null;
var g_workChunk = 0;

var g_computeTime = 0;
var g_pass = 0;
var g_exploring = {row960: "", moves: []};
var g_longestAttempt = {row960: "", moves: []};
var g_solutions = [];
var g_solutionsNum = 0;	// Different from g_solutions.length if "count only".

var g_nodesBatch = 0;
var g_nodesCount = 0;
var g_nodesComputeTime = 0;
var g_nodesSpeed = "?";

function setFirstPlyAndInitialMoveNumber() {
	g_firstPly = 0;
	g_initialMoveNumber = 1;

	// First pass to look at the first play or stipulation command.
	for (var i = 0; i < g_commands.length; i++) {
		if (g_commands[i].command == "play")
			break;
		else if (g_commands[i].command == "stipulation") {
			var subcommands = g_commands[i].subcommands;
			for (var j = 0; j < subcommands.length; j++) {
				if (subcommands[j].command != "option")
					continue;
				if (subcommands[j].type.value == "FirstToMove") {
					if (subcommands[j].color.value == "black")
						g_firstPly = 1;
				}
				else if (subcommands[j].type.value == "InitialMoveNumber")
					g_initialMoveNumber = subcommands[j].number.value;
			}
			break;
		}
	}

	// Second pass to look at options, which have priority.
	for (var i = 0; i < g_commands.length; i++) {
		if (g_commands[i].command == "option") {
			if (g_commands[i].type.value == "FirstToMove") {
				if (g_commands[i].color.value == "black")
					g_firstPly = 1;
				else
					g_firstPly = 0;
			}
			else if (g_commands[i].type.value == "InitialMoveNumber")
				g_initialMoveNumber = g_commands[i].number.value;
		}
	}
}

function updateProgressStats(fraction, load) {
	// Update speed estimate every 5 second.
	var newBatch = Math.floor(g_computeTime / 5000);
	if (newBatch != g_nodesBatch) {
		g_nodesSpeed = (g_nodesCount / g_nodesComputeTime * 1000).toFixed(0);
		g_nodesCount = 0;
		g_nodesComputeTime = 0;
		g_nodesBatch = newBatch;
	}

//	if (fraction == 1 || g_nodesCount == 0) {
	var passName = (g_pass == 0) ? "" : "pass " + g_pass.toString() + ": ";
	var percent = Math.floor(fraction * 10000) / 100;
	var info = passName + percent.toFixed(2) + "%, " +
	           Math.floor(g_computeTime / 1000).toFixed(0) + " s, " +
	           g_nodesSpeed + " nodes/sec";
	if (DEBUG) {
		var percent2 = Math.floor(load * 10000) / 100;
		info += ", " + percent2.toFixed(2) + "% load";
	}
	$("span#percent").html("(" + info + ")");

//	}
}

function appendTestResult(fraction) {
	// Write test reports.
	if (fraction == 1)
		g_testCompleteTime += g_computeTime;
	else
		g_testIncompleteTime += g_computeTime;

	$("div#test-results").append(
		g_tests[g_testIndex][1] + " " +
		fraction.toPrecision(4) + " " +
		(g_computeTime / 1000).toFixed(3) + " " +
		g_solutionsNum + " " +
		(g_testCompleteTime / 1000).toFixed(3) + " " +
		(g_testIncompleteTime / 1000).toFixed(3) + "<br>");

	$("div#test-results-box").show();
}

function findPieceIndex(c) {
	var src = pieceLetters.englishN;	// Notation returned by the engine.
	for (var i = 0; i < src.length; i++) {
		if (c == src[i])
			return i;
	}
	return -1;
}

function translateMoveToFigurine(move, color) {
	var dest = (color == "w") ? pieceFigurines.w : pieceFigurines.b;

	var ret = "";
	// Remember index of last piece seen. We need it if we encounter
	// "=w" or "=b" not followed by a piece.
//	var last_index = 14;	// Use undefined piece for Masand color.
	var last_index = 0;	// Assume pawn since it has no letter.

	for (var j = 0; j < move.length; ) {
		var c = move.charAt(j);
		if (c == '[')
			last_index = 14;	// Forget last piece seen because Masand.
		if (c == "=" || c == "+") {
			ret += c;
			j++;

			c = move.charAt(j);
			// Can only check for these right after a "=" or a "+" because
			// 'b' is also a file.
			if (c == "w" || c == "b") {
				var index = findPieceIndex(move.charAt(j+1));
				if (index >= 0) {
					ret += pieceFigurines[c][index];
					last_index = index;
					j += 2;
				}
				else {
					ret += pieceFigurines[c][last_index];
					j++;
				}
			}
		}
		else {
			var index = findPieceIndex(c);
			if (index >= 0) {
				ret += dest[index];
				last_index = index;
			}
			else
				ret += c;
			j++;
		}
	}

	return ret;
}

function solutionToHtml(soln, lang) {
	var letters;
	var colors;
	if (lang == "englishN") {
		letters = pieceLetters.englishN;
		colors = ["w", "b"];
	}
	else if (lang == "englishS" || lang == "py2web") {
		letters = pieceLetters.englishS;
		colors = ["w", "b"];
	}
	else if (lang == "french" || lang == "french2") {
		letters = pieceLetters.french;
		colors = ["b", "n"];
	}
	else if (lang == "german") {
		letters = pieceLetters.german;
		colors = ["w", "s"];
	}
	else if (lang == "figurine") {
		letters = pieceFigurines.w;
		colors = [pieceFigurines.w[14], pieceFigurines.b[14]];			
	}
	else
		alert("Error: unknown output language '" + lang + "'");

	var ret = "";
	if (soln.row960 != "") {
		ret += "(";
		for (var i = 0; i < soln.row960.length; i++) {
			var c = soln.row960.charAt(i);
			var index = findPieceIndex(c);
			if (index >= 0)
				ret += letters[index];
			else
				ret += c;
		}
		ret += ") ";
	}
	for (var i = 0; i < soln.moves.length; i++) {
		if (i > 0 && (numberTwice % 2 != 0 || soln.moves[i-1].san != ""))
			ret += " ";

		var numberTwice = i + g_initialMoveNumber*2;
		if (numberTwice % 2 == 0) {
			ret += (numberTwice/2).toString() + ".";
		}
		else {
			if (i == 0) {
				ret += ((numberTwice - 1)/2).toString() + ".";
				if (soln.moves[i].san != "")
					ret += "..";
			}
			else {
				if (soln.moves[i].san != "" && soln.moves[i-1].san == "")
					ret += "..";
			}
		}

		if (lang == "py2web")
			ret += soln.moves[i].py + soln.moves[i].disamb;
		else if (lang == "englishN")
			ret += soln.moves[i].san;
		else if (lang == "figurine") {
			ret += translateMoveToFigurine(soln.moves[i].san,
			                               soln.moves[i].color);
		}
		else {
			var move = soln.moves[i].san;

			for (var j = 0; j < move.length; j++) {
				var c = move.charAt(j);
				if (c == "=") {
					if (move.charAt(j+1) == "w") {
						ret += "=" + colors[0];
						j++;
						continue;
					}
					else if (move.charAt(j+1) == "b") {
						ret += "=" + colors[1];
						j++;
						continue;
					}
				}
				var index = findPieceIndex(c);
				if (index >= 0)
					ret += letters[index];
				else if (lang == "french2" && c == "c")
					ret += "&#231;";
				else if (lang == "french2" && c == "e")
					ret += "&#233;";
				else
					ret += c;
			}
		}
	}

	return ret;
}

// In the returned solution, the first ply is stored at index 0 even if
// Black plays first.
function extractSolution(depth) {
	var ret = {row960: Solver_GetRow960(),
	           moves: []};
	for (var i = g_firstPly; i < depth; i++)
		SP_ReverseMove(0);
	for (var i = g_firstPly; i < depth; i++) {
		SP_PlayMove(0, -1);
		ret.moves.push({san: SP_GetLastMoveNotation(0, 0),
		                py: SP_GetLastMoveNotation(0, 1),
		                disamb: SP_GetLastMoveNotation(0, 2),
		                color: SP_GetLastMoveNotation(0, 3)});
	}
	return ret;
}

function solutionIsEqual(soln, soln2) {
	if (soln.row960 != soln2.row960)
		return false;
	for (var j = 0; j < soln.moves.length; j++) {
		if (soln.moves[j].san != soln2.moves[j].san)
			return false;
	}
	return true;
}

function solutionIsNew(soln) {
	for (var i = 0; i < g_solutions.length; i++) {
		if (solutionIsEqual(soln, g_solutions[i]))
			return false;
	}
	return true;
}

function catchFn(e) {
	if (e.name || e.message) {
		// Javascript error, like "Too much recursion".
		$("div#crash-text").append(e.name + ": " + e.message);
	}
	else {
		if (!$("div#crash-box").is(":visible")) {
			// abort() was called but nothing has been written to stdout or
			// stderr. This only happens in the following cases I think.
			$("div#crash-text").append("Error in malloc, free, or realloc.");
		}

		// "e" is probably an Emscripten stack trace.
		// Only print it in debug mode.
		if (DEBUG)
			$("div#crash-text").append(e);
	}

	$("div#crash-box").show();

	g_paused = true;
	$("button#pause-resume").addClass("ui-state-disabled");

	// We only need this for Javascript errors, but it doesn't hurt
	// to set it always in case our test above is wrong.
	ABORT = true;
}

function runSolver() {
	var startTime = (new Date()).getTime();

	if (g_lastCurrTime == null)
		g_lastCurrTime = startTime;

	if (g_paused) {
		g_lastCurrTime = null;	// No setTimeout in flight.
		return;
	}

	var lang = $("select#output-language").val();

	var currTime;
	var depth;
	var initialStack = Module["asm"].stackSave();

	try {
		// The try-catch will catch both Javascript errors like
		// "too much recursion" and C calls to abort().
		// Many functions could potentially crash:
		// Solver_RunStep, Solver_IsPassIncomplete, Solver_PrepareDemolitionPass
		// extractSolution.

	while (true) {
		depth = Solver_RunStep();
		g_nodesCount++;

		if (depth == g_firstPly) {
			if (g_pass > 0) {
				if (Solver_IsPassIncomplete()) {
					g_pass++;
					Solver_PrepareDemolitionPass(Math.pow(0.95, g_pass));
					continue;
				}
			}

			// Search is over!
			g_paused = true;
			$("button#pause-resume").addClass("ui-state-disabled");

			currTime = (new Date()).getTime();
			g_computeTime += currTime - g_lastCurrTime;
			g_nodesComputeTime += currTime - g_lastCurrTime;
			updateProgressStats(1, TT_LoadFraction());

			// Signal to tell not to translate text.
			g_exploring = {row960: "", moves: []};
			$("div#exploring").html("Done. Compute time = " +
			                        (g_computeTime / 1000).toFixed(3) + " s");
			if (g_solutionsNum == 0)
				$("span#soln_text").text("No exact solution. Longest attempt");

			g_lastCurrTime = null;	// No setTimeout in flight.

			if (g_testIndex < g_tests.length) {
				appendTestResult(1);
				if (++g_testIndex < g_tests.length)
					runCurrTest();	// Launch another search.
			}
			return;
		}

		if (depth == g_lastPly) {	// We found a solution.
			var flood = $("select#flood-protection").val();
			// count-only is incompatible with demolition mode.
			if (g_pass > 0 && flood == "count-only")
				flood = "auto-pause";

			if (g_solutionsNum >= 100 && flood == "count-only") {
				g_solutionsNum++;
				$("span#soln_text").text("exact solutions (" +
			                        	 g_solutionsNum + ")");
			}
			else {
				// Add solution.
				var soln = extractSolution(depth);
				if (g_pass == 0 || solutionIsNew(soln)) {
					g_solutions.push(soln);
					g_solutionsNum++
					$("span#soln_text").text("exact solutions (" +
				                        	 g_solutionsNum + ")");
					if (g_solutionsNum == 1) {
						// Erase longest attempt.
						$("div#solutions").html("<ul>\n</ul>");
					}
					$("div#solutions ul").append(
						"<li>" + solutionToHtml(soln, lang) + "</li>");

					if (flood == "auto-pause") {
						var n = g_solutionsNum;
						if (n == 100 || n == 1000 || n == 10000 ||
						    n == 100000 || n == 1000000) {
							g_paused = true;
							currTime = (new Date()).getTime();
							alert("The solver has been paused because " + 
							      "it found " +  n + " solutions. " +
								  "Click 'Resume' to search for more.");
							break;
						}
					}
				}
			}
		}
		else if (g_solutionsNum == 0 &&
		         depth - g_firstPly > g_longestAttempt.moves.length) {
			g_longestAttempt = extractSolution(depth);
			$("div#solutions").html(
				solutionToHtml(g_longestAttempt, lang));
		}

		currTime = (new Date()).getTime();
		// Chrome issues a "Violation" when a handler takes more than 50 ms,
		// so try to stay below that limit.
		if (currTime - startTime > 40)	// 40 ms.
			break;
	}

	g_computeTime += currTime - g_lastCurrTime;
	g_nodesComputeTime += currTime - g_lastCurrTime;

	g_workChunk++;
	// Do not update UI on every work chunk to save some CPU. 
	if (g_workChunk % 3 == 0) {
		updateProgressStats(Solver_SearchFraction(), TT_LoadFraction());

		g_exploring = extractSolution(depth);
		$("div#exploring").html(
			solutionToHtml(g_exploring, lang));
	}

	}	// Closing try.
	catch (e) {
		catchFn(e);
		// This partially resets Emscripten to make it more stable when
		// we run it again. Is there a way to reset it completely?
		Module["asm"].stackRestore(initialStack);
		g_lastCurrTime = null;	// No setTimeout in flight.
		return;
	}

	g_lastCurrTime = currTime;

	if (g_testIndex < g_tests.length && g_computeTime / 1000 >= g_testTimeout) {
		g_lastCurrTime = null;	// No setTimeout in flight.

		// Time out this test.
		appendTestResult(Solver_SearchFraction());
		if (++g_testIndex < g_tests.length)
			runCurrTest();	// Launch the next test.
		else {
			// Pause because the last test has timed out.
			g_paused = true;
		}
		return;
	}

	// setTimeout doesn't work at full performance because
	// browsers replace the 0 with a minimum value of 4 ms.
	// See https://www.nczonline.net/blog/2013/07/09/the-case-for-setimmediate/
	setImmediate(runSolver);
//	setTimeout(runSolver, 0);
}

function buildConditionCheckboxes(list, selectorPrefix) {
	$fieldsetLeft = $(selectorPrefix + "-left");
	$fieldsetRight = $(selectorPrefix + "-right");
	var html = "";
	var lines = 0;

	for (var i = 0; i < list.length; i++) {
		var parts = list[i].split(": ");
		var name = parts[0];
		var index = name.indexOf("_");

		if (index == -1 && i > 0) {
			// Flush html block.
			var blockMidpoint = i - lines/2;

			if (blockMidpoint <= list.length / 2)
				$fieldsetLeft.append(html);
			else
				$fieldsetRight.append(html);
			html = "";
			lines = 0;
		}

		var shortName = false;
		var displayName;
		if (index == -1)
			displayName = name;
		else {
			var name2 = name.substr(index + 1);
			if (name2.length <= 2)
				shortName = true;
			displayName = "&#8212;&#160;" + name2;
			// Good looking, but not in Firefox for Android.
//			displayName = "&#10551;&#160;" + name.substr(index + 1);
			// Too distracting.
//			displayName = "&#8627;&#160;" + name.substr(index + 1);
		}
		if (shortName)
			html += '<div style="display:inline-block; vertical-align:top;">'
		if (index == -1)
			html += '<label>'
		else
			html += '<label class="level2">'	// Currently has no effect.
		html += '<input id="' + name + '" type="checkbox">' +
		        displayName + '</label>\n';
		if (shortName)
			html += '</div>'
		lines++;
	}

	// Flush last block.
	$fieldsetRight.append(html);
}

function refreshUIConditions() {
	$("div#conditions input").prop("checked", false);

	for (var i = 0; i < g_commands.length; i++) {
		if (g_commands[i].command == "condition") {
			var conditions = g_commands[i].conditions;
			for (var j = 0; j < conditions.length; j++) {
				// Click the elementary conditions.
				var parts = conditions[j].value.split(" ");
				for (var k = 0; k < parts.length; k++) {
					$("div#conditions input#" + parts[k]).prop(
						"checked", true);
				}
				// Click the combo conditions.
				for (var k = 0; k < comboNames.length; k++) {
					var parts = comboNames[k].split(": ");
					if (conditions[j].value == parts[1]) {
						$("div#conditions input#" + parts[0]).prop(
							"checked", true);
						break;
					}
				}
			}
		}
	}
	$("div#conditions input").checkboxradio("refresh");
}

function refreshTextArea() {
// https://stackoverflow.com/questions/10699210/jquery-mobile-growing-text-area-with-ajax-data
	$("textarea#text-representation").val(textFromCommands(g_commands))
	                                 .keyup();
}

// https://stackoverflow.com/questions/901115/how-can-i-get-query-string-values-in-javascript
function getParameterByName(name) {
	// Escape any bracket in name.
	name = name.replace(/[\[\]]/g, "\\$&");
 	var regex = new RegExp("[?&]" + name + "(=([^&#]*)|&|#|$)");
	var results = regex.exec(location.search);
	if (!results) return null;
	if (!results[2]) return "";
	return decodeURIComponent(results[2].replace(/\+/g, " "));
}

// Only call these when the UI is in sync with g_commands.

function addAntiParryIfNeeded() {
	var $AP = $("div#conditions input#AntiParry");
	var hasAntiParry = $AP.is(":checked");

	var needAntiParry = false;
	for (var i = 0; i < g_commands.length; i++) {
		if (g_commands[i].command == "stipulation") {
			var prefix = g_commands[i].prefix.value;
			if (prefix == "ahser-" || prefix == "pahser-" ||
			    prefix == "ah" || prefix == "pah") {
				needAntiParry = true;
			}
		}
	}

	if (needAntiParry && !hasAntiParry) {
		$AP.prop("checked", true)
		   .checkboxradio("refresh")
	       .trigger("change");
		alert("Warning: This problem contains the stipulation prefix " +
		      "ahser-, pahser-, ah or pah " +
			  "without the fairy condition AntiParry. " +
			  "The fairy condition has been automatically added.");
	}
}

function addAutoChecksIfNeeded() {
	var $AP = $("div#conditions input#AntiParry");
	var hasAntiParry = $AP.is(":checked");
	var $ATAC = $("div#conditions input#AllowTerminalAutoChecks");
	var hasATAC = $ATAC.is(":checked");

	var testingAutoCheck = false;
	var testingAutoMate = false;
	// Scan tests and stipulations.
	for (var i = 0; i < g_commands.length; i++) {
		if (g_commands[i].command == "test" ||
		    g_commands[i].command == "set" ||
		    g_commands[i].command == "stipulation") {
			var aim = g_commands[i].aim.value;
			if (aim == parseInt(findInList("!+", testNames)))
				testingAutoCheck = true;
			else if (aim == parseInt(findInList("!#", testNames)) ||
			         aim == parseInt(findInList("##", testNames))) {
				testingAutoMate = true;
			}
		}
	}

	if (testingAutoMate && !hasATAC) {
		$ATAC.prop("checked", true)
		     .checkboxradio("refresh")
		     .trigger("change");
		alert("Warning: This problem is testing for !# or ## without a fairy " +
		      "condition that can generate auto-mates. The condition " +
			  "AllowTerminalAutoChecks has been automatically added.");
	}
	else if (testingAutoCheck && !hasATAC && !hasAntiParry) {
		$ATAC.prop("checked", true)
		     .checkboxradio("refresh")
			 .trigger("change");
		alert("Warning: This problem is testing for !+ without a fairy " +
		      "condition that can generate auto-checks. The condition " +
		      "AllowTerminalAutoChecks has been automatically added.");
	}
}

function addRoyalDynastyIfNeeded() {
	var $RD = $("div#conditions input#RoyalDynasty");

	if ($("div#conditions input#NonRoyalKings").is(":checked") ||
	    $("div#conditions input#Extinction").is(":checked") ||
		$RD.is(":checked")) {
		return;
	}

	var cond;
	if ($("div#conditions input#Knightmate").is(":checked")) {
		if (!$("div#conditions input#CavalierMajeur").is(":checked") &&
		    $("div#conditions input#PromOnly_S").is(":checked")) {
			cond = "Knightmate + PromOnly S";
		}
		else
			return;	// PromOnly N doesn't exist yet, so we're fine.
	}
	else { // Knightmate is off.
		if ($("div#conditions input#PromOnly_K").is(":checked"))
			cond = "PromOnly K";
		else
			return;
	}

	$RD.prop("checked", true)
	   .checkboxradio("refresh")
       .trigger("change");
	alert("Warning: " + cond + " cannot be used without one " +
	      "of NonRoyalKings, Extinction, or RoyalDynasty. " +
	      "RoyalDynasty has been automatically added.");
}

function removeCommand(commands, i) {
	commands.splice(i, 1);
	if (i == 0 && commands.length > 0) {
		// Remove any leading whitespace.
		while (isWhiteSpace(commands[0].text.charAt(0)))
			commands[0].text = commands[0].text.substr(1);
	}
}

function parseMemory(str) {
	var regex = /^[ \t\n\r]*([0-9eE.]*)[ \t\n\r]*(|kB|K|MB|M|GB|G)[ \t\n\r]*$/;
	var results = regex.exec(str);
	if (results == null)
		return NaN;
	var mem = parseFloat(results[1]);

	// Allow the same suffixes as Unix dd:
	mem *= {"": 1,
	        "kB": 1000, "K": 1024,
	        "MB": 1000*1000, "M": 1024*1024,
	        "GB": 1000*1000*1000, "G": 1024*1024*1024}[results[2]];
	return mem;
}

function startSolving() {
	if (!g_engineReady) {
		// Try again in 10ms.
		setTimeout(startSolving, 10);
		return;
	}

	// Clear UI now because we could crash soon.
	g_longestAttempt = {row960: "", moves: []};
	g_solutions = [];
	g_solutionsNum = 0;

	g_computeTime = 0;

	g_nodesBatch = 0;
	g_nodesCount = 0;
	g_nodesComputeTime = 0;
	g_nodesSpeed = "?";

	$("span#percent").html("");
	$("div#exploring").html("");
	$("span#soln_text").text("No exact solution so far. Longest attempt");
	$("div#solutions").html("");

	setFirstPlyAndInitialMoveNumber();
	var demolitionMode = $("input#demolition-mode").is(":checked");

	ABORT = false;	// Clear any previous crash to be able to run again.
	var initialStack = Module["asm"].stackSave();
	try {
		SetProblemFromJSON(JSON.stringify(g_commands), g_firstPly,
		                   demolitionMode);
	}
	catch (e) {
		catchFn(e);
		Module["asm"].stackRestore(initialStack);
		return;
	}

	if (Solver_PawnOnIllegalRank()) {
		alert("There is a pawn on rank 1 or 8 which is impossible with the " +
		      "given conditions. If you have a good reason to want this, " +
		      "then tell me and maybe I'll add an option.");
		return;
	}

	try {
		// This function can crash, for example with:
		//   pieces white Ka4 Rh1 black Kh4 Ra8
		//   stip h~0.5
		//   cond VerticalCylinder BackHome
		if (Solver_CanCaptureTheKing()) {
			alert("The side to play can capture the king");
			return;
		}
	}
	catch (e) {
		catchFn(e);
		Module["asm"].stackRestore(initialStack);
		alert("A crash occurred while determining if the side to play can " +
		      "capture the king.");
		return;
	}
	
	g_lastPly = GetLastPly();
	g_paused = false;

	if (demolitionMode) {
		g_pass = 1;
		try {
			Solver_PrepareDemolitionPass(0.95);
		}
		catch (e) {
			catchFn(e);
			Module["asm"].stackRestore(initialStack);
			return;
		}
	}
	else
		g_pass = 0;	// Ordinary mode.

	$("button#pause-resume").removeClass("ui-state-disabled");

	if (g_lastCurrTime == null)
		runSolver();
	// Else we already have a setTimeout in flight.
}

$(document).one("pagebeforecreate", "#solver", function() {
	buildConditionCheckboxes(conditionNames, "fieldset#conditions");
	buildConditionCheckboxes(comboNames, "fieldset#combos");

	// Somehow the theme doesn't apply everywhere. Help JQuery Mobile get the
	// job done.
	var theme = $("div#solver").attr("data-theme");
	$("div.ui-bar").addClass("ui-bar-" + theme);
	$("div.ui-body").addClass("ui-body-" + theme);
});

$(document).one("pagecreate", "#solver", function() {
	var $page = $(this);	// "this" is the page for pagecreate.

	// Is the AppCache implementation of Firefox and/or Chrome severely
	// broken even in a simple case? The code below detects non-uniform
	// updates.
	var $version = $page.find("div[data-role='header'] h1");
	if ($version.html() != "Jacobi " + g_version)
		$version.append(" (misloaded, JavaScript is " + g_version + ")");

	$page.find("select#first-to-move").on("change", function(event, ui) {
		var colorE = $(event.target).val();

		// Color name in chosen language.
		var color = "";
		if (colorE == "white")
			color = pieceColorNames[g_language][0];
		else if (colorE == "black")
			color = pieceColorNames[g_language][1];

		var sep = color.indexOf(":");
		if (sep >= 0)
			color = color.substr(0, sep);

		var seen = 0;
		var insert_index = -1;

		// Traverse the commands to either delete the option or select
		// a good place to insert it.
		for (var i = 0; i < g_commands.length; i++) {
			if (g_commands[i].command == "option") {
				if (g_commands[i].type.value == "FirstToMove") {
					seen++;
					if (seen >= 2 || colorE == "unset") {
						removeCommand(g_commands, i);
						i--;
						continue;
					}
					else if (seen == 1) {
						g_commands[i].color =
							{text: " " + color, value: colorE};
					}
				}
			}
			// Just before these commands would be a good spot.
			if (insert_index == -1 && (
			    (g_commands[i].command == "option" &&
					g_commands[i].type.value == "InitialMoveNumber") ||
			    g_commands[i].command == "play" ||
			    g_commands[i].command == "test" ||
			    g_commands[i].command == "set" ||
				g_commands[i].command == "stipulation")) {
				insert_index = i;
			}
		}

		// At the end would be a good spot.
		if (insert_index == -1)
			insert_index = g_commands.length;

		if (colorE != "unset" && seen == 0) {
			g_commands.splice(insert_index, 0, {
				text: (insert_index == 0 ? "" : "\n") + "option",
				command: "option",
				type: {text: " FirstToMove", value: "FirstToMove"},
				color: {text: " " + color, value: colorE}
			});
			if (insert_index == 0 && g_commands.length >= 2)
				g_commands[1].text = "\n" + g_commands[1].text;
		}

		refreshTextArea();
	});

	$page.find("input#initial-move-number").on("change", function(event, ui) {
		var token = $(event.target).val();
		var re = /^-?\d*(|\.0|\.5)$/;
		if (!re.test(token)) {
			alert("Error: initial move number '" + token + "' is wrong.");
			return;
		}
		var number = parseFloat(token);

		var seen = 0;
		var insert_index = -1;

		// Traverse the commands to either delete the option or select
		// a good place to insert it.
		for (var i = 0; i < g_commands.length; i++) {
			if (g_commands[i].command == "option") {
				if (g_commands[i].type.value == "InitialMoveNumber") {
					seen++;
					if (seen >= 2 || token == "") {
						removeCommand(g_commands, i);
						i--;
						continue;
					}
					else if (seen == 1) {
						g_commands[i].number =
							{text: " " + token, value: number};
					}
				}
			}
			// Just before these commands would be a good spot.
			if (insert_index == -1 && (
			    g_commands[i].command == "play" ||
			    g_commands[i].command == "test" ||
			    g_commands[i].command == "set" ||
				g_commands[i].command == "stipulation")) {
				insert_index = i;
			}
		}

		// At the end would be a good spot.
		if (insert_index == -1)
			insert_index = g_commands.length;

		if (number != 0 && seen == 0) {
			g_commands.splice(insert_index, 0, {
				text: (insert_index == 0 ? "" : "\n") + "option",
				command: "option",
				type: {text: " InitialMoveNumber", value: "InitialMoveNumber"},
				number: {text: " " + token, value: number}
			});
			if (insert_index == 0 && g_commands.length >= 2)
				g_commands[1].text = "\n" + g_commands[1].text;
		}

		refreshTextArea();
	});

	$page.find("div#conditions input").on("change", function(event, ui) {
		var name = event.target.id;
		var checked = event.target.checked;

		// Is this a combo condition?
		combo = "";
		for (var k = 0; k < comboNames.length; k++) {
			var parts = comboNames[k].split(": ");
			if (name == parts[0]) {
				combo = parts[1];
				break;
			}
		}

		var inserted = false;
		var insert_index = -1;

		var affectedBases = [];
		// Traverse the commands to either delete the condition or select
		// a good place to insert it.
		for (var i = 0; i < g_commands.length; i++) {
			if (g_commands[i].command == "condition") {
				var conditions = g_commands[i].conditions;
				var insert_subindex = -1;
				for (var j = 0; j < conditions.length; j++) {
					var cond = conditions[j].value;

					if (cond == combo) {
						// The user clicked a combo button and it matches
						// this condition.
						if (!checked) {	// We want it removed.
							conditions.splice(j, 1);
							j--;
						}
						else {
							if (DEBUG) {
								console.log("'" + combo +
								            "' is already in commands");
							}
						}
						continue;
					}

					var prefixSet = [];
					addConditionBases(prefixSet, cond);
					if (prefixSet.indexOf(conditionBase(name)) < 0)
						continue;	// Not matching.

					addConditionBases(affectedBases, cond);

					var parts = cond.split(" ");
					if (parts.length > 1 &&
					    (!checked || prefixSet.length > 1)) {
						// We need to break the special condition apart.
						conditions.splice(j, 1);
						var j2 = j;
						for (var k = 0; k < parts.length; k++) {
							conditions.splice(j2, 0, {
								text: " " + conditionTail(parts[k]),
						        value: parts[k]});
							j2++;
						}
						// Simpler to reprocess conditions[j].
						j--;
						continue;
					}

					if (!checked) {
						if (name == cond || name == conditionBase(cond)) {
							// We know the condition has only one part.
							conditions.splice(j, 1);
							j--;
						}
					}
					else {
						// Since there was a base match, this spot would work.
						insert_subindex = j+1;	// This spot would work.
					}
				}

				if (insert_subindex >= 0) {
					conditions.splice(insert_subindex, 0, {
						text: " " + conditionTail(name),
						value: name});
					inserted = true;
					break;
				}

				if (conditions.length == 0) {
					// Get rid of empty command.
					removeCommand(g_commands, i);
					i--;
				}
				insert_index = i+1;
			}
			else if (g_commands[i].command != "ignore") {
				// Just after the last command that we don't ignore would be a
				// good spot.
				insert_index = i+1;
			}
		}

		// At the end would be a good spot.
		if (insert_index == -1)
			insert_index = g_commands.length;

		if (checked && !inserted) {
			var conditions;
			if (combo != "")
				conditions = [{text: " " + name, value: combo}];
			else if (name == conditionBase(name))
				conditions = [{text: " " + name, value: name}];
			else {
				conditions = [{text: " " + conditionBase(name),
				               value: conditionBase(name)},
				              {text: " " + conditionTail(name),
							   value: name}];
			}

			g_commands.splice(insert_index, 0, {
				text: (g_commands.length == 0 ? "" : "\n") + "condition",
				command: "condition",
				conditions: conditions
			});
		}

		// TODO: see if the conditions with an affected base name can be joined
		// in a combo from comboNames.

		// Refresh conditions from scratch. Otherwise if we do it by delta and
		// there's a bug, then the UI will be out-of-sync.
		refreshUIConditions();

		refreshTextArea();
	});

	$page.find("button#apply-changes").on("vclick", function(event, ui) {
		// Parse textarea.
		var text = $("textarea#text-representation").val();

		// Split the content into "separators + tokens" parts.
		var parts = [];
		var i = 0;
		while (i < text.length) {
			var pair = {separator: "", token: ""};
			// Grab separator.
			while (i < text.length) {
				var c = text.charAt(i);
				if (!isWhiteSpace(c))
					break;
				pair.separator += c;
				i++
			}
			// Grab token.
			while (i < text.length) {
				var c = text.charAt(i);
				if (isWhiteSpace(c))
					break;
				pair.token += c;
				i++
			}
			parts.push(pair);
		}
//		console.log(parts);

		// Parse parts into commands.
		var commands = [];
		var parser = {
			parts: parts,
			curr: 0,
			language: "englishS",

			getCurrToken: function() {
				return (this.curr < this.parts.length) ?
				       this.parts[this.curr].token : "";
			},

			getCurrSeparator: function() {
				return this.parts[this.curr].separator;
			}
		};

		var success = true;
		while (parser.curr < parser.parts.length) {
			success = parseCommand(parser, commands);
			if (!success)
				break;
		}
		if (DEBUG) {
			console.log("success = " + success + ", commands = ", commands);
			console.log(textFromCommands(commands));
			console.log(JSON.stringify(commands));
		}
		if (!success)
			return;

		// Update options in the UI. First, assume defaults.
		$("select#first-to-move").val("unset");
		$("input#initial-move-number").val("");

		for (var i = 0; i < commands.length; i++) {
			if (commands[i].command == "option") {
				if (commands[i].type.value == "FirstToMove") {
					$("select#first-to-move").val(commands[i].color.value);
//						Not needed: JQuery does the job.
//						.attr('selected', true)
//						.siblings('option').removeAttr('selected');
				}
				else if (commands[i].type.value == "InitialMoveNumber") {
					$("input#initial-move-number").val(
						commands[i].number.value);
				}
			}
		}
		$("select#first-to-move").selectmenu("refresh", true);
		$("div#conditions input").checkboxradio("refresh");

		g_commands = commands;
		g_language = parser.language;

		refreshUIConditions();
	});

	$page.find("button#start").on("vclick", function(event, ui) {
		$("div#crash-box").hide();
		$("div#crash-text").html("<p>Congratulations, you managed to crash Jacobi! Consider sending a bug report to flab&#64;wismuth&#46;com to get this particular bug fixed. Please copy the problem you were solving and the details below.</p>");

		addAntiParryIfNeeded();
		addAutoChecksIfNeeded();
		addRoyalDynastyIfNeeded();

		if (!g_engineReady) {
			if (g_engineLaunched)
				return;	// User clicked "start" twice. Ignore second click.

			var str = $("input#hash-mem").val();
			g_hashMem = parseMemory(str);
			if (isNaN(g_hashMem)) {
				alert("Error: Hash table memory '" + str +
				      "' was not understood.");
					  return;
			}
			$("input#hash-mem").addClass("ui-state-disabled");

			g_useWasm = typeof WebAssembly == "object" &&
			            getParameterByName("asmjs") == null;

			// Load the engine asynchronously.
			var script = document.createElement("script");
			var dir = g_version.split(" ")[0];
			var prefix = "../" + dir + "/engine";
			script.src = prefix + (g_useWasm ? "-wasm.js" : "-asmjs.js");
			document.body.appendChild(script);
			g_engineLaunched = true;
		}

		startSolving();
	});

	$page.find("button#pause-resume").on("vclick", function(event, ui) {
		if (g_paused) {
			g_paused = false;
			runSolver();	
		}
		else
			g_paused = true;
	});

	$page.find("select#output-language").on("change", function(event, ui) {
		// Rewrite solutions with new language.
		var lang = $(event.target).val();

		if (g_exploring.moves.length > 0) {	// Do not replace "DONE."
			$("div#exploring").html(
				solutionToHtml(g_exploring, lang));
		}

		if (g_solutionsNum == 0) {
			$("div#solutions").html(
				solutionToHtml(g_longestAttempt, lang));
		}
		else {
			var html = "";
			for (var i = 0; i < g_solutions.length; i++)
				html += "<li>" + solutionToHtml(g_solutions[i], lang) +
				        "</li>";
			$("div#solutions ul").html(html);
		}
	});
});

// ---------------------------------------------------------------------------

// One em = 16px if !wantMini().
// One em = 12.5px if wantMini().

function wantMini() {
	return ($(window).width() < 560);
}

// Call this initially and on resize.
function adjustSizes($scope) {
	// Can't set elements. Have to set an enclosing tag.
	$scope.find("div.ui-content").toggleClass("ui-mini", wantMini());
	$scope.find("div[data-role='header']").toggleClass("ui-mini", wantMini());
	$scope.find("div[data-role='controlgroup']").toggleClass(
		"ui-mini", wantMini());

	// Fix label font-size which doesn't adjust.
	$scope.find("label").css(
		"font-size", wantMini() ? "12.5px" : "16px");

	$scope.find("ul.ui-listview").toggleClass("ui-mini", wantMini());
	$scope.find("fieldset").toggleClass("ui-mini", wantMini());
}

$(window.document).on("pagebeforecreate", function(event, ui) {
	// Don't use $.mobile.activePage which refers to the page we're
	// transitioning from.
	var $page = $(event.target);

	// Must be done on pagebeforecreate otherwise it doesn't work.
	// Sizes end up being fixed after that and don't adjust when resizing.
	adjustSizes($page);
});

$(window).resize(function() {
	adjustSizes($("body"));
});

// ---------------------------------------------------------------------------

var g_tests = [];
var g_testIndex = 0;
var g_testTimeout = 300;
var g_testCompleteTime;
var g_testIncompleteTime;

function runCurrTest() {
	$("textarea#text-representation").val(g_tests[g_testIndex][2])
	                                 .keyup();
	$("button#apply-changes").trigger("vclick");
	$("button#start").trigger("vclick");
}

// For local Chrome,
// run "chromium --allow-file-access-from-files" and solver.html?asmjs .
function runTests(url) {
	$.ajax({
		url: url,
		dataType: "xml",
		success: function(data) {			
			$(data).find("problem").each(function() {
				var text = $(this).text();
				var lines = text.split("\n");
				if (lines[0] != "")
					alert("missing newline");
				var tokens = lines[1].split(" ");
				if (tokens[0] != "origin")
					alert("missing origin:\n" + text);

				g_tests.push([Infinity, tokens[1], text]);
			});
			var results = $(data).find("results").text().split("\n");

			// Attach results' time to the corresponding test.
			for (var i = 0; i < results.length; i++) {
				var parts = results[i].split(" ");
				if (parts.length == 6) {
					var origin = parts[0];
					var time = parseFloat(parts[2]) / parseFloat(parts[1]);
					var found = false;
					for (var j = 0; j < g_tests.length; j++) {
						if (g_tests[j][1] == origin) {
							if (found)
								alert("repeated origin: " + origin);
							g_tests[j][0] = time;
						}
					}
				}
			}

			// Reorder to solve the faster problems first.
			g_tests.sort(function(a,b) { return a[0] - b[0]; });

			// Estimated time before first incomplete solution.
			var i;
			var sum = 0;
			for (i = 0; i < g_tests.length; i++) {
				var t = g_tests[i][0];
				if (t > g_testTimeout)
					break;
				sum += t;
			}
			console.log("estimated " + i + " problems in " + sum.toFixed(0) +
			            " s before first incomplete solution.");
			console.log("estimated " + g_tests.length + " problems in " +
			            (sum + (g_tests.length - i)*g_testTimeout).toFixed(0) +
			            " s total.");

			g_testIndex = 0;
			g_testCompleteTime = 0;
			g_testIncompleteTime = 0;
			$("select#flood-protection").val("count-only")
			                            .selectmenu("refresh", true);
			runCurrTest();
		},
		error: function(jqXHR, textStatus, errorThrown) {
			console.log("ajax error: " + textStatus + ", " + errorThrown);
		}
	});
}

// ---------------------------------------------------------------------------
